#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Dict, List, Optional

import trafaret as t

from datarobot.models.api_object import APIObject


class ExternalBaselineValidationInfo(APIObject):
    """An object containing information about external time series baseline predictions
    validation results.

    Attributes
    ----------
    baseline_validation_job_id : str
        the identifier of the baseline validation job
    project_id : str
        the identifier of the project
    catalog_version_id : str
        the identifier of the catalog version used in the validation job
    target : str
        the name of the target feature
    datetime_partition_column : str
        the name of the column whose values as dates are used to assign a row
        to a particular partition
    is_external_baseline_dataset_valid : bool
        whether the external baseline dataset passes the validation check
    multiseries_id_columns : List[str] or null
        a list of the names of multiseries id columns to define series
        within the training data.  Currently only one multiseries id column is supported.
    holdout_start_date : str or None
        the start date of holdout scoring data
    holdout_end_date : str or None
        the end date of holdout scoring data
    backtests : list of dicts containing validation_start_date and validation_end_date or None
        the configured backtests of the time series project
    forecast_window_start : int
        offset into the future to define how far forward relative to the forecast point the
        forecast window should start.
    forecast_window_end : int
        offset into the future to define how far forward relative to the forecast point the
        forecast window should end.
    message : str or None
        the description of the issue with external baseline validation job

    """

    _get_url = "projects/{pid}/externalTimeSeriesBaselineDataValidationJobs/{job_id}/"

    _converter = t.Dict(
        {
            t.Key("baseline_validation_job_id"): t.String(),
            t.Key("project_id"): t.String(),
            t.Key("catalog_version_id"): t.String(),
            t.Key("target"): t.String(),
            t.Key("datetime_partition_column"): t.String(),
            t.Key("is_external_baseline_dataset_valid"): t.Bool(),
            t.Key("multiseries_id_columns", optional=True): t.List(
                t.String(), max_length=1, min_length=1
            ),
            t.Key("holdout_start_date", optional=True): t.String(),
            t.Key("holdout_end_date", optional=True): t.String(),
            t.Key("backtests", optional=True): t.List(
                t.Dict(
                    {
                        t.Key("validation_start_date"): t.String(),
                        t.Key("validation_end_date"): t.String(),
                    }
                )
            ),
            t.Key("forecast_window_start", optional=True): t.Int(),
            t.Key("forecast_window_end", optional=True): t.Int(),
            t.Key("message", optional=True): t.String(),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        baseline_validation_job_id: str,
        project_id: str,
        catalog_version_id: str,
        target: str,
        datetime_partition_column: str,
        is_external_baseline_dataset_valid: bool,
        multiseries_id_columns: Optional[List[str]] = None,
        holdout_start_date: Optional[str] = None,
        holdout_end_date: Optional[str] = None,
        backtests: Optional[List[Dict[str, str]]] = None,
        forecast_window_start: Optional[int] = None,
        forecast_window_end: Optional[int] = None,
        message: Optional[str] = None,
    ) -> None:
        self.baseline_validation_job_id = baseline_validation_job_id
        self.project_id = project_id
        self.catalog_version_id = catalog_version_id
        self.target = target
        self.datetime_partition_column = datetime_partition_column
        self.is_external_baseline_dataset_valid = is_external_baseline_dataset_valid
        self.multiseries_id_columns = multiseries_id_columns
        self.holdout_start_date = holdout_start_date
        self.holdout_end_date = holdout_end_date
        self.backtests = backtests
        self.forecast_window_start = forecast_window_start
        self.forecast_window_end = forecast_window_end
        self.message = message

    @classmethod
    def get(cls, project_id: str, validation_job_id: str) -> ExternalBaselineValidationInfo:
        """
        Get information about external baseline validation job

        Parameters
        ----------
        project_id : string
            the identifier of the project
        validation_job_id : string
            the identifier of the external baseline validation job

        Returns
        -------
        info: ExternalBaselineValidationInfo
            information about external baseline validation job

        """
        server_data = cls._client.get(cls._get_url.format(pid=project_id, job_id=validation_job_id))
        return cls.from_server_data(server_data.json())



# flake8: noqa
from .augmentation import ImageAugmentationList, ImageAugmentationOptions, ImageAugmentationSample
from .images import DuplicateImage, Image, SampleImage
from .insights import ImageActivationMap, ImageEmbedding



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Union

from datarobot.models.genai.comparison_chat import get_entity_id
from datarobot.models.genai.insights_configuration import InsightsConfiguration
from datarobot.models.genai.playground import Playground


class MetricInsights(InsightsConfiguration):
    """Metric insights for playground."""

    _path = "api/v2/genai/playgrounds/{}/supportedInsights/"

    @classmethod
    def list(cls, playground: Union[str, Playground]) -> list[InsightsConfiguration]:
        """Get metric insights for playground.

        Parameters
        ----------
        playground : str or Playground
            Playground to get the supported metrics from.

        Returns
        -------
        insights: list[InsightsConfiguration]
            Metric insights for playground.

        """
        path = cls._path.format(get_entity_id(playground))

        response_data = cls._client.get(url=f"{cls._client.domain}/{path}")
        return [
            cls.from_server_data(insight)
            for insight in response_data.json()["insightsConfiguration"]
        ]

    @classmethod
    def copy_to_playground(
        cls,
        source_playground: Union[str, Playground],
        target_playground: Union[str, Playground],
        add_to_existing: bool = True,
        with_evaluation_datasets: bool = False,
    ) -> None:
        """Copy metric insights to from one playground to another.

        Parameters
        ----------
        source_playground : str or Playground
            Playground to copy metric insights from.
        target_playground : str or Playground
            Playground to copy metric insights to.
        add_to_existing : Optional[bool]
            Add metric insights to existing ones in the target playground, by default True.
        with_evaluation_datasets : Optional[bool]
            Copy evaluation datasets from the source playground.
        """
        payload = {
            "add_to_existing": add_to_existing,
            "with_evaluation_datasets": with_evaluation_datasets,
        }
        path = (
            cls._path.format(get_entity_id(target_playground))
            + f"{get_entity_id(source_playground)}/"
        )
        cls._client.put(f"{cls._client.domain}/{path}", json=payload)



#
# Copyright 2021-2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

from __future__ import annotations

import os
from typing import Any, cast, Dict, Optional, TYPE_CHECKING, Union

import trafaret as t
from urllib3 import Retry
import yaml

from ._compat import Int, String
from .context import ENABLE_API_CONSUMER_TRACKING_DEFAULT, env_to_bool
from .rest import DataRobotClientConfig

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    class ConfigDict(TypedDict, total=False):
        """TypedDict for config"""

        token: Optional[str]
        endpoint: Optional[str]
        config_path: Optional[str]
        connect_timeout: Optional[int]
        ssl_verify: Optional[bool]
        max_retries: Optional[Union[int, Retry]]
        token_type: Optional[str]
        default_use_case: Optional[str]
        enable_api_consumer_tracking: Optional[bool]
        trace_context: Optional[str]
        use_tcp_keepalive: Optional[bool]


_file_exists = os.path.isfile

_converter = t.Dict(
    {
        t.Key("endpoint"): String(),
        t.Key("token"): String(),
        t.Key("connect_timeout", optional=True): Int(),
        t.Key("ssl_verify", optional=True): t.Or(t.Bool(), String()),
        t.Key("max_retries", optional=True): Int(),
        t.Key("token_type", optional=True): String(),
        t.Key("default_use_case", optional=True): String(),
        t.Key("enable_api_consumer_tracking", optional=True): t.Bool(),
        t.Key("trace_context", optional=True): String(),
        t.Key("use_tcp_keepalive", optional=True): t.Bool(),
    }
).allow_extra("*")

_fields = {k.to_name or k.name for k in _converter.keys}


def _get_first_non_none_value(*args: Any) -> Union[Any, None]:
    return next((arg for arg in args if arg is not None), None)


# Custom function to get the first non-none value and its source
def _get_value_and_source(*values: Any) -> Union[tuple[Any, int], tuple[None, None]]:
    return next(
        ((value, index) for index, value in enumerate(values) if value is not None), (None, None)
    )


def _config_dict_from_data(data: Dict[str, Any]) -> Dict[str, Any]:
    return {k: v for k, v in _converter.check(data).items() if k in _fields}


def _get_config_file_from_env() -> Optional[str]:
    if "DATAROBOT_CONFIG_FILE" in os.environ:
        config_path = os.environ["DATAROBOT_CONFIG_FILE"]
        expanded_config_path = os.path.expanduser(config_path)
        if os.path.exists(expanded_config_path):
            return expanded_config_path
        else:
            raise ValueError("Environment variable DATAROBOT_CONFIG_FILE points to a missing file")
    return None


def _get_config_dir() -> str:
    return os.path.expanduser("~/.config/datarobot")


def _get_default_config_file() -> Optional[str]:
    first_choice_config_path = os.path.join(_get_config_dir(), "drconfig.yaml")
    if _file_exists(first_choice_config_path):
        return first_choice_config_path
    else:
        return None


def _config_from_env() -> ConfigDict:
    """
    Create and return a DataRobotClientConfig from environment variables. This method only allows
    for configuration of endpoint, token, max retries, and a default use case.
    More advanced configuration must be done through a yaml file.

    Returns
    -------
    config : ConfigDict

    Raises
    ------
    ValueError
        If either of DATAROBOT_ENDPOINT or DATAROBOT_API_TOKEN is not specified as an environment
        variable
    IOError
        If the config file that DATAROBOT_CONFIG_FILE points to does not exist
    """
    endpoint: Optional[str] = os.environ.get("DATAROBOT_ENDPOINT")
    token: Optional[str] = os.environ.get("DATAROBOT_API_TOKEN")
    ssl_verify: Optional[Union[str, bool]] = os.environ.get("DATAROBOT_SSL_VERIFY")
    if ssl_verify is not None:
        ssl_verify = env_to_bool(ssl_verify)
    trace_context: Optional[str] = os.environ.get("DATAROBOT_TRACE_CONTEXT")
    max_retries: Optional[Union[str, int]] = os.environ.get("DATAROBOT_MAX_RETRIES")
    if max_retries is not None:
        max_retries = int(max_retries)
    use_case_id: Optional[str] = os.environ.get("DATAROBOT_DEFAULT_USE_CASE")
    enable_api_consumer_tracking: Optional[Union[str, bool]] = os.environ.get(
        "DATAROBOT_API_CONSUMER_TRACKING_ENABLED"
    )
    if enable_api_consumer_tracking is not None:
        enable_api_consumer_tracking = env_to_bool(enable_api_consumer_tracking)
    use_tcp_keepalive: Optional[Union[str, bool]] = os.environ.get("DATAROBOT_USE_TCP_KEEPALIVE")
    if use_tcp_keepalive is not None:
        use_tcp_keepalive = env_to_bool(use_tcp_keepalive)

    return cast(
        "ConfigDict",
        {
            "endpoint": endpoint,
            "token": token,
            "ssl_verify": ssl_verify,
            "max_retries": max_retries,
            "default_use_case": use_case_id,
            "enable_api_consumer_tracking": enable_api_consumer_tracking,
            "trace_context": trace_context,
            "connect_timeout": None,
            "use_tcp_keepalive": use_tcp_keepalive,
        },
    )


def _config_from_file(config_path: str) -> ConfigDict:
    """
    Create and return a DataRobotClientConfig from a config path. The file must be
    a yaml formatted file

    Parameters
    ----------
    config_path : str
        Path to the configuration file

    Returns
    -------
    config : ConfigDict
    """
    with open(config_path, "rb") as f:
        data: Dict[str, Any] = yaml.load(f, Loader=yaml.SafeLoader)
    return cast("ConfigDict", _config_dict_from_data(data))


# Create client config section
def create_drconfig(
    token: Optional[str] = None,
    endpoint: Optional[str] = None,
    config_path: Optional[str] = None,
    connect_timeout: Optional[int] = None,
    ssl_verify: Optional[bool] = None,
    max_retries: Optional[Union[int, Retry]] = None,
    token_type: Optional[str] = None,
    default_use_case: Optional[str] = None,
    enable_api_consumer_tracking: Optional[bool] = None,
    trace_context: Optional[str] = None,
    use_tcp_keepalive: Optional[bool] = None,
) -> DataRobotClientConfig:
    """
    Create a DataRobotClientConfig object from the provided arguments, from environment
    variables or a config file.

    Notes
    -----
    Token and endpoint must be specified from one source only. This is a restriction
    to prevent token leakage if environment variables or config file are used.

    The DataRobotClientConfig params will be looking up to find the configuration parameters
    in one of the following ways,

      1. From call kwargs if specified;
      2. From a YAML file at the path specified in the ``config_path`` kwarg;
      3. From a YAML file at the path specified in the environment variables ``DATAROBOT_CONFIG_FILE``;
      4. From environment variables;
      5. From the default values in the default YAML file
         at the path `$HOME/.config/datarobot/drconfig.yaml`.

    Parameters
    ----------
    token : Optional[str]
        API token.
    endpoint : Optional[str]
        Base URL of API.
    config_path : Optional[str]
        An alternate location of the config file.
    connect_timeout : Optional[int]
        How long the client should be willing to wait before giving up on establishing
        a connection with the server.
    ssl_verify : bool or Optional[str]
        Whether to check SSL certificate.
        Could be set to path with certificates of trusted certification authorities. Default: True.
    max_retries : int or urllib3.util.retry.Retry, optional
        Either an integer number of times to retry connection errors,
        or a `urllib3.util.retry.Retry` object to configure retries.
    token_type: Optional[str]
        Authentication token type: Token, Bearer.
        "Bearer" is for DataRobot OAuth2 token, "Token" for token generated in Developer Tools.
        Default: "Token".
    default_use_case: Optional[str]
        The entity ID of the default Use Case to use with any requests made by this Client instance.
    enable_api_consumer_tracking: Optional[bool]
        Enable and disable user metrics tracking within the datarobot module. Default: False.
    trace_context: Optional[str]
        An ID or other string for identifying which code template or AI Accelerator was used to make
        a request.
    use_tcp_keepalive : Optional[bool]
        Enable TCP keepalive for the client. If None is set to True. Default: None.

    Returns
    -------
    config : DataRobotClientConfig
    """

    # Validate that if endpoint is specified, token is also specified
    if endpoint and not token:
        raise ValueError("Token must be specified if endpoint is specified")

    env_config = _get_config_file_from_env()
    # If a config path is specified, use that
    if config_path:
        config_path_expanded = os.path.expanduser(config_path)
        if not _file_exists(config_path_expanded):
            raise ValueError(f"Invalid config path - no file at {config_path_expanded}")
        config = _config_from_file(config_path_expanded)
    # If no config path is specified, check for an env var
    elif env_config:
        if not _file_exists(env_config):
            raise ValueError(f"Invalid config path - no file at {env_config}")
        config = _config_from_file(env_config)
    # If no config path is specified, check for a environmet variables config file
    else:
        config = _config_from_env()
    default_config = cast("ConfigDict", {})
    default_config_path = _get_default_config_file()
    if default_config_path is not None:
        default_config = _config_from_file(default_config_path)
    if not config.get("max_retries", None):
        config["max_retries"] = Retry(backoff_factor=0.1, respect_retry_after_header=True)

    token, token_source = _get_value_and_source(
        token, config.get("token"), default_config.get("token")
    )
    endpoint, endpoint_source = _get_value_and_source(
        endpoint, config.get("endpoint"), default_config.get("endpoint")
    )
    # Raise an error if no endpoint or token is specified
    if endpoint is None or token is None:
        e_msg = (
            "No valid configuration found"
            "enpoint and token must be specified"
            "Can be specified via arguments, environment variables or a config file"
        )
        raise ValueError(e_msg)

    # Raise an error if token or endpoint are from different sources
    # This should prevent token leakage
    if token_source != endpoint_source:
        raise ValueError("Endpoint and token must come from the same configuration source")

    ssl_verify = _get_first_non_none_value(
        ssl_verify, config.get("ssl_verify"), default_config.get("ssl_verify")
    )

    max_retries = _get_first_non_none_value(
        max_retries, config.get("max_retries"), default_config.get("max_retries")
    )

    connect_timeout = _get_first_non_none_value(
        connect_timeout,
        config.get("connect_timeout"),
        default_config.get("connect_timeout"),
    )
    token_type = _get_first_non_none_value(
        token_type, config.get("token_type"), default_config.get("token_type", "Token")
    )
    default_use_case = _get_first_non_none_value(
        default_use_case,
        config.get("default_use_case"),
        default_config.get("default_use_case"),
    )
    enable_api_consumer_tracking = _get_first_non_none_value(
        enable_api_consumer_tracking,
        config.get("enable_api_consumer_tracking"),
        default_config.get("enable_api_consumer_tracking", ENABLE_API_CONSUMER_TRACKING_DEFAULT),
    )
    trace_context = _get_first_non_none_value(
        trace_context, config.get("trace_context"), default_config.get("trace_context")
    )

    use_tcp_keepalive = _get_first_non_none_value(
        use_tcp_keepalive, config.get("use_tcp_keepalive"), default_config.get("use_tcp_keepalive")
    )

    return DataRobotClientConfig(
        token=token,
        endpoint=endpoint,
        ssl_verify=ssl_verify if ssl_verify is not None else True,
        max_retries=max_retries,
        connect_timeout=connect_timeout,
        token_type=token_type,
        default_use_case=default_use_case,
        enable_api_consumer_tracking=enable_api_consumer_tracking,
        trace_context=trace_context,
        use_tcp_keepalive=use_tcp_keepalive if use_tcp_keepalive is not None else True,
    )



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, NamedTuple, Optional, Type, TypeVar

from requests import Response
import trafaret as t

from datarobot._compat import String
from datarobot.utils import from_api

from ..client import get_client, staticproperty
from ..enums import ASYNC_PROCESS_STATUS, DEFAULT_MAX_WAIT
from ..errors import AsyncFailureError, AsyncProcessUnsuccessfulError, AsyncTimeoutError
from ..utils.waiters import wait_for_async_resolution
from .api_object import APIObject

TStatusCheckJob = TypeVar("TStatusCheckJob", bound="StatusCheckJob")
"""
This class represents job for status check of submitted async jobs.
This class is not in AbstractJob hierarchy because of differences in api "/status/" call method.
"""


class StatusCheckJob:
    """Tracks asynchronous task status

    Attributes
    ----------
    job_id : str
        The ID of the status the job belongs to.
    """

    _client = staticproperty(get_client)  # type: ignore[arg-type]

    _converter_common = t.Dict(
        {
            t.Key("status", optional=True): t.Enum(
                ASYNC_PROCESS_STATUS.ABORTED,
                ASYNC_PROCESS_STATUS.COMPLETED,
                ASYNC_PROCESS_STATUS.ERROR,
                ASYNC_PROCESS_STATUS.INITIALIZED,
                ASYNC_PROCESS_STATUS.RUNNING,
            ),
            t.Key("statusId", optional=True): String,
        }
    )

    def __init__(self, job_id: str, resource_type: Optional[Type[APIObject]] = None) -> None:
        """
        Pass in resource type to be used for `self.get_result_when_complete`

        job_id : str
            The ID of the status the job belongs to
        resource_type : APIObject
            The type of the resource expected to be returned once the job has completed.  This is used to
            automatically create a resource of the appropriate type once the final resource is available.
        """
        self.job_id = job_id
        self.converter = self._converter_common.allow_extra("*")
        self.resource_type = resource_type

    def status_from_response(
        self, data: Dict[str, Any], completed_resource_url: Optional[str] = None
    ) -> JobStatusResult:
        safe_data = self.converter.check(from_api(data))

        return JobStatusResult(
            status=safe_data.get("status"),
            status_id=safe_data.get("statusId"),
            completed_resource_url=completed_resource_url,
            message=safe_data.get("message"),
        )

    @classmethod
    def _job_path(cls, job_id: str) -> str:
        return f"status/{job_id}/"

    def _this_job_path(self) -> str:
        return self._job_path(self.job_id)

    @classmethod
    def from_id(cls: Type[TStatusCheckJob], job_id: str) -> StatusCheckJob:
        return cls(job_id)

    @classmethod
    def from_response(
        cls: Type[TStatusCheckJob],
        response: Response,
        response_type: Optional[Type[APIObject]] = None,
    ) -> StatusCheckJob:
        location_string = response.headers["Location"]
        job_id = location_string.split("/")[-2]
        return cls(job_id, response_type)

    def wait_for_completion(self, max_wait: int = DEFAULT_MAX_WAIT) -> JobStatusResult:
        """
        Waits for job to complete.

        Parameters
        ----------
        max_wait : Optional[int]
            How long to wait for the job to finish. If the time expires, DataRobot returns the current status.

        Returns
        -------
        status : JobStatusResult
            Returns the current status of the job.
        """
        try:
            wait_for_async_resolution(self._client, self._this_job_path(), max_wait=max_wait)
        except (AsyncFailureError, AsyncProcessUnsuccessfulError) as ex:
            return self.status_from_response(
                data={"status": ASYNC_PROCESS_STATUS.ERROR, "message": str(ex)}
            )
        except AsyncTimeoutError:
            pass  # just return current status to user

        return self.get_status()

    def get_status(self) -> JobStatusResult:
        """
        Retrieve JobStatusResult object with the latest job status data from the server.
        """
        response = self._client.get(self._this_job_path(), allow_redirects=False)

        if response.status_code == 200:
            data = response.json()
            return self.status_from_response(data=data)
        elif response.status_code == 303:
            completed_url = response.headers["Location"]
            return self.status_from_response(
                data={"status": ASYNC_PROCESS_STATUS.COMPLETED},
                completed_resource_url=completed_url,
            )
        else:
            e_msg = "Server unexpectedly returned status code {}"
            raise AsyncFailureError(e_msg.format(response.status_code))

    def get_result_when_complete(self, max_wait: int = DEFAULT_MAX_WAIT) -> APIObject:
        """
        Wait for the job to complete, then attempt to convert the resulting json into an object of type
        self.resource_type
        Returns
        -------
        A newly created resource of type self.resource_type
        """

        if not self.resource_type:
            raise ValueError("The function requires self.resource_type to be set before calling")

        # if this fails to complete let it throw an exception
        async_result = wait_for_async_resolution(
            self._client, self._this_job_path(), max_wait=max_wait
        )

        # if we got here we should be complete, fetch the resource
        resource = (
            async_result
            if isinstance(async_result, str)
            else async_result["completed_resource_url"]
        )
        return self.resource_type.from_location(resource)


class JobStatusResult(NamedTuple):
    status: Optional[str]
    status_id: Optional[str]
    completed_resource_url: Optional[str]
    message: Optional[str]



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from abc import ABCMeta
from collections import defaultdict

import numpy as np
import pandas as pd
import trafaret as t

from datarobot._compat import Int, String
from datarobot.models.api_object import APIObject

VALID_RELEVANCE_CONFIGURATIONS = {(0, 0), (0, 1), (1, 0), (1, 1)}


class PairwiseStatisticsBase(
    APIObject, metaclass=ABCMeta
):  # pylint: disable=missing-class-docstring
    _converter = t.Dict(
        {
            t.Key("feature_name"): String,
            t.Key("data"): t.List(
                t.Dict(
                    {
                        t.Key("label_configuration"): t.List(
                            t.Dict(
                                {t.Key("relevance", optional=True): Int, t.Key("label"): String}
                            ),
                            min_length=2,
                            max_length=2,
                        ),
                        # In case of missing values we will have None here, which gets removed in
                        # from_location call (keep_atters can not be used because data is a list)
                        t.Key("statistic_value", optional=True): t.Float,
                    }
                ).allow_extra("*")
            ),
        }
    ).allow_extra("*")

    def __init__(self, feature_name, data):
        self.feature_name = feature_name
        self.values = data

    def __repr__(self):
        return f"{self.__class__.__name__}(feature_name={self.feature_name})"

    @classmethod
    def _get(cls, multilabel_insights_key, statistic_type):
        url = "multilabelInsights/{}/pairwiseStatistics/?statisticType={}".format(
            multilabel_insights_key, statistic_type
        )
        return cls.from_location(url)

    @staticmethod
    def _sort_index_and_columns(df):
        """
        Sorts the index and columns of a pairwise statistics dataframe.
        The dataframe is expected to have a shape of (num_labels, num_labels) and to have the same
        label_names in the index and the columns.
        """
        df = df.sort_index()
        df = df[df.index]
        return df


class PairwiseCorrelations(PairwiseStatisticsBase):
    """
    Correlation of label pairs for multicategorical feature.

    .. versionadded:: v2.24

    Notes
    -----
    ``CorrelationValues`` contain:

    * ``values.[].label_configuration`` : list of length 2 - Configuration of the label pair
    * ``values.[].label_configuration.[].label`` : str – Label name
    * ``values.[].statistic_value`` : float – Statistic value

    Attributes
    ----------
    feature_name : str
        Name of the feature
    values : list(dict)
        List of correlation values with a schema described as ``CorrelationValues``
    statistic_dataframe : pandas.DataFrame
        Correlation values for all label pairs as a DataFrame
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.statistic_dataframe = self._to_statistic_dataframe(self.values)

    @staticmethod
    def _to_statistic_dataframe(correlation_values):
        """
        Converts the correlation values to a dataframe.

        Parameters
        ----------
        correlation_values : list(dict)
            List of correlation values following the ``CorrelationValues`` schema outlined in the
            class description.

        Returns
        -------
        pandas.DataFrame
            The correlation values as a pandas.DataFrame
        """
        columns = defaultdict(dict)
        for statistic_value in correlation_values:
            row_label = str(statistic_value["label_configuration"][0]["label"])
            column_label = str(statistic_value["label_configuration"][1]["label"])
            columns[column_label][row_label] = statistic_value.get("statistic_value", np.nan)

        statistic_dataframe = pd.DataFrame(columns)
        statistic_dataframe = PairwiseCorrelations._sort_index_and_columns(statistic_dataframe)

        return statistic_dataframe

    @classmethod
    def get(cls, multilabel_insights_key):
        """Retrieves pairwise correlations

        You might find it more convenient to use
        :meth:`Feature.get_pairwise_correlations
        <datarobot.models.Feature.get_pairwise_correlations>`
        instead.

        Parameters
        ----------
        multilabel_insights_key: string
            Key for multilabel insights, unique for a project, feature and EDA stage combination.
            The multilabel_insights_key can be retrieved via
            ``Feature.multilabel_insights_key``.

        Returns
        -------
        PairwiseCorrelations
            The pairwise label correlations
        """
        return cls._get(multilabel_insights_key, "correlation")

    def as_dataframe(self):
        """The pairwise label correlations as a (num_labels x num_labels) DataFrame.

        Returns
        -------
        pandas.DataFrame
            The pairwise label correlations. Index and column names allow the interpretation of the
            values.
        """
        return self.statistic_dataframe


class PairwiseProbabilitiesBase(
    PairwiseStatisticsBase, metaclass=ABCMeta
):  # pylint: disable=missing-class-docstring
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.statistic_dataframes = self._to_statistic_dataframes(self.values)

    @staticmethod
    def _to_statistic_dataframes(statistic_values):
        """
        Converts the statistic values into DataFrames. There will be one DataFrame for each
        relevance configuration.

        Parameters
        ----------
        statistic_values : list(dict)
            List of statistic values following the ``ProbabilityValues`` schema outlined in the
            class description of PairwiseJointProbabilities or PairwiseConditionalProbabilities.

        Returns
        -------
        dict(pandas.DataFrame)
            The keys of the returned dictionary are the relevance configurations
            e.g. (0, 0), (0, 1) etc. The values are the static_values for this relevance
            configuration as pandas.DataFrame.
        """
        grouped_statistic_values = defaultdict(lambda: defaultdict(dict))
        for statistic_value in statistic_values:
            relevance_configuration = (
                statistic_value["label_configuration"][0]["relevance"],
                statistic_value["label_configuration"][1]["relevance"],
            )
            row_label = str(statistic_value["label_configuration"][0]["label"])
            column_label = str(statistic_value["label_configuration"][1]["label"])
            grouped_statistic_values[relevance_configuration][column_label][row_label] = (
                statistic_value.get("statistic_value", np.nan)
            )
        statistic_dataframes = {}
        for relevance_configuration, columns in grouped_statistic_values.items():
            df = pd.DataFrame(columns)
            df = PairwiseProbabilitiesBase._sort_index_and_columns(df)
            statistic_dataframes[relevance_configuration] = df
        return statistic_dataframes

    def _as_dataframe(self, relevance_configuration):
        if relevance_configuration not in VALID_RELEVANCE_CONFIGURATIONS:
            raise ValueError(
                "You have passed an invalid label configuration. "
                "Valid options are (0, 0), (0, 1), (1, 0) and (1, 1)"
            )
        return self.statistic_dataframes[relevance_configuration]


class PairwiseJointProbabilities(PairwiseProbabilitiesBase):
    """
    Joint probabilities of label pairs for multicategorical feature.

    .. versionadded:: v2.24

    Notes
    -----
    ``ProbabilityValues`` contain:

    * ``values.[].label_configuration`` : list of length 2 - Configuration of the label pair
    * ``values.[].label_configuration.[].relevance`` : int – 0 for absence of the labels,
      1 for the presence of labels
    * ``values.[].label_configuration.[].label`` : str – Label name
    * ``values.[].statistic_value`` : float – Statistic value

    Attributes
    ----------
    feature_name : str
        Name of the feature
    values : list(dict)
        List of joint probability values with a schema described as ``ProbabilityValues``
    statistic_dataframes : dict(pandas.DataFrame)
        Joint Probability values as DataFrames for different relevance combinations.

        E.g. The probability P(A=0,B=1) can be retrieved via:
        ``pairwise_joint_probabilities.statistic_dataframes[(0,1)].loc['A', 'B']``
    """

    @classmethod
    def get(cls, multilabel_insights_key):
        """Retrieves pairwise joint probabilities

        You might find it more convenient to use
        :meth:`Feature.get_pairwise_joint_probabilities
        <datarobot.models.Feature.get_pairwise_joint_probabilities>`
        instead.

        Parameters
        ----------
        multilabel_insights_key: string
            Key for multilabel insights, unique for a project, feature and EDA stage combination.
            The multilabel_insights_key can be retrieved via
            ``Feature.multilabel_insights_key``.

        Returns
        -------
        PairwiseJointProbabilities
            The pairwise joint probabilities
        """
        return cls._get(multilabel_insights_key, "jointProbability")

    def as_dataframe(self, relevance_configuration):
        """Joint probabilities of label pairs as a (num_labels x num_labels) DataFrame.

        Parameters
        ----------
        relevance_configuration: tuple of length 2
            Valid options are (0, 0), (0, 1), (1, 0) and (1, 1). Values of 0 indicate absence of
            labels and 1 indicates presence of labels. The first value describes the
            presence for the labels in axis=0 and the second value describes the presence for the
            labels in axis=1.

            For example the matrix values for a relevance configuration of (0, 1) describe the
            probabilities of absent labels in the index axis and present labels in the column
            axis.

            E.g. The probability P(A=0,B=1) can be retrieved via:
            ``pairwise_joint_probabilities.as_dataframe((0,1)).loc['A', 'B']``

        Returns
        -------
        pandas.DataFrame
            The joint probabilities for the requested ``relevance_configuration``. Index and column
            names allow the interpretation of the values.
        """
        return self._as_dataframe(relevance_configuration)


class PairwiseConditionalProbabilities(PairwiseProbabilitiesBase):
    """
    Conditional probabilities of label pairs for multicategorical feature.

    .. versionadded:: v2.24

    Notes
    -----
    ``ProbabilityValues`` contain:

    * ``values.[].label_configuration`` : list of length 2 - Configuration of the label pair
    * ``values.[].label_configuration.[].relevance`` : int – 0 for absence of the labels,
      1 for the presence of labels
    * ``values.[].label_configuration.[].label`` : str – Label name
    * ``values.[].statistic_value`` : float – Statistic value

    Attributes
    ----------
    feature_name : str
        Name of the feature
    values : list(dict)
        List of conditional probability values with a schema described as ``ProbabilityValues``
    statistic_dataframes : dict(pandas.DataFrame)
        Conditional Probability values as DataFrames for different relevance combinations.
        The label names in the columns are the events, on which we condition. The label names in the
        index are the events whose conditional probability given the indexes is in the dataframe.

        E.g. The probability P(A=0|B=1) can be retrieved via:
        ``pairwise_conditional_probabilities.statistic_dataframes[(0,1)].loc['A', 'B']``
    """

    @classmethod
    def get(cls, multilabel_insights_key):
        """Retrieves pairwise conditional probabilities

        You might find it more convenient to use
        :meth:`Feature.get_pairwise_conditional_probabilities
        <datarobot.models.Feature.get_pairwise_conditional_probabilities>`
        instead.

        Parameters
        ----------
        multilabel_insights_key: string
            Key for multilabel insights, unique for a project, feature and EDA stage combination.
            The multilabel_insights_key can be retrieved via
            ``Feature.multilabel_insights_key``.

        Returns
        -------
        PairwiseConditionalProbabilities
            The pairwise conditional probabilities
        """
        return cls._get(multilabel_insights_key, "conditionalProbability")

    def as_dataframe(self, relevance_configuration):
        """Conditional probabilities of label pairs as a (num_labels x num_labels) DataFrame.
        The label names in the columns are the events, on which we condition. The label names in the
        index are the events whose conditional probability given the indexes is in the dataframe.

        E.g. The probability P(A=0|B=1) can be retrieved via:
        ``pairwise_conditional_probabilities.as_dataframe((0, 1)).loc['A', 'B']``

        Parameters
        ----------
        relevance_configuration: tuple of length 2
            Valid options are (0, 0), (0, 1), (1, 0) and (1, 1). Values of 0 indicate absence of
            labels and 1 indicates presence of labels. The first value describes the
            presence for the labels in axis=0 and the second value describes the presence for the
            labels in axis=1.

            For example the matrix values for a relevance configuration of (0, 1) describe the
            probabilities of absent labels in the index axis given the
            presence of labels in the column axis.

        Returns
        -------
        pandas.DataFrame
            The conditional probabilities for the requested ``relevance_configuration``.
            Index and column names allow the interpretation of the values.
        """
        return self._as_dataframe(relevance_configuration)



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import List, Optional

import trafaret as t

from datarobot._compat import String
from datarobot.models.api_object import APIObject

from ..utils.pagination import unpaginate


class PredictionServer(APIObject):
    """A prediction server can be used to make predictions.

    Attributes
    ----------
    id : Optional[str]
        The id of the prediction server.
    url : str
        The url of the prediction server.
    datarobot_key : Optional[str]
        The ``Datarobot-Key`` HTTP header used in requests to this prediction server. Note that in the
        :class:`datarobot.models.Deployment` instance there is the ``default_prediction_server``
        property which has this value as a "kebab-cased" key as opposed to "snake_cased".
    """

    _path = "predictionServers/"
    _converter = t.Dict(
        {
            t.Key("id", optional=True) >> "id": String(),
            t.Key("url") >> "url": String(allow_blank=True),
            t.Key("datarobot-key", optional=True) >> "datarobot_key": String(allow_blank=True),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: Optional[str] = None,
        url: Optional[str] = None,
        datarobot_key: Optional[str] = None,
    ) -> None:
        self.id = id
        self.url = url
        self.datarobot_key = datarobot_key

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.url or self.id})"

    @classmethod
    def list(cls) -> List[PredictionServer]:
        """Returns a list of prediction servers a user can use to make predictions.

        .. versionadded:: v2.17

        Returns
        -------
        prediction_servers : list of PredictionServer instances
            Contains a list of prediction servers that can be used to make predictions.

        Examples
        --------
        .. code-block:: python

            prediction_servers = PredictionServer.list()
            prediction_servers
            >>> [PredictionServer('https://example.com')]
        """

        data = unpaginate(cls._path, {}, cls._client)
        return [cls.from_server_data(item) for item in data]



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

"""
"Overall" configuration for a model is not specific to a particular moderation guard.
Typically these are pushed, along with guard configurations, to a new custom model version.

Settings are:

    timeout_sec: the maximum time in seconds allowed for moderation
    timeout_action: what do to if moderation times out: block or continue
    entity_id: the object (such as custom model version) this config applies to
    entity_type: the type of the related entity
    updated_at: the time this configuration was updated (not manually settable)
    updater_id: the ID of the user making the update (not manually settable)
"""

from __future__ import annotations

from typing import Any, Dict, Optional, Set

import trafaret as t

from datarobot import errors
from datarobot.enums import ModerationGuardEntityType, ModerationTimeoutActionType
from datarobot.models.api_object import APIObject
from datarobot.utils.pagination import unpaginate

OVERALL_MODERATION_DEFAULT_TIMEOUT_SEC = 60


class OverallModerationConfig(APIObject):
    """Details of overall moderation configuration for a model."""

    _path = "overallModerationConfiguration/"

    _converter = t.Dict(
        {
            t.Key("timeout_sec"): t.Int(),
            t.Key("timeout_action"): t.Enum(*[e.value for e in ModerationTimeoutActionType]),
            t.Key("entity_id"): t.String(),
            t.Key("entity_type"): t.Enum(*[e.value for e in ModerationGuardEntityType]),
            t.Key("updated_at", optional=True): t.String(),
            t.Key("updater_id", optional=True): t.String(),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(self, **kwargs: Any) -> None:
        self._set_values(**kwargs)

    def __repr__(self) -> str:
        return "Overall moderation configuration; see details"

    def _set_values(
        self: OverallModerationConfig,
        timeout_sec: int,
        timeout_action: ModerationTimeoutActionType,
        entity_id: str,
        entity_type: ModerationGuardEntityType,
        updated_at: str = "",
        updater_id: str = "",
    ) -> None:
        """
        Initialize an overall configuration object based on values from init or server data.
        """
        self.timeout_sec = timeout_sec
        self.timeout_action = ModerationTimeoutActionType(timeout_action)
        self.entity_id = entity_id
        self.entity_type = ModerationGuardEntityType(entity_type)
        self.updated_at = updated_at
        self.updater_id = updater_id

    def to_dict(self: OverallModerationConfig) -> Dict[str, str | int]:
        return {
            "timeoutSec": self.timeout_sec,
            "timeoutAction": self.timeout_action,
            "entityId": self.entity_id,
            "entityType": self.entity_type,
            "updatedAt": self.updated_at,
            "updaterId": self.updater_id,
        }

    @classmethod
    def _template_path(cls, template_id: str) -> str:
        return f"{cls._path}{template_id}/"

    def _update_values(
        self: OverallModerationConfig, new_response: OverallModerationConfig
    ) -> None:
        # called by update() and refresh()
        fields: Set[str] = self._fields()  # type: ignore[no-untyped-call]
        for attr in fields:
            new_value = getattr(new_response, attr)
            setattr(self, attr, new_value)

    @classmethod
    def find(
        cls, entity_id: str, entity_type: ModerationGuardEntityType
    ) -> Optional[OverallModerationConfig]:
        """Find overall configuration by entity ID and entity type.
        Each entity (such as a customModelVersion) may have at most 1 overall moderation configuration.

        .. versionadded:: v3.6

        Parameters
        ----------
        entity_id: str
            ID of the entity
        entity_type: str
            Type of the entity

        Returns
        -------
        Optional[OverallModerationConfig]
            an OverallModerationConfig or None

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        data = unpaginate(
            cls._path,
            {"entityId": entity_id, "entityType": entity_type.value},
            cls._client,
        )
        result = next(data, None)
        if not result:
            return None
        return cls.from_server_data(result)

    @classmethod
    def locate(
        cls, entity_id: str, entity_type: ModerationGuardEntityType
    ) -> OverallModerationConfig:
        """Find overall configuration by entity ID and entity type.
        This version of find() expects the object to exist. Its return type is not optional.

        .. versionadded:: v3.6

        Parameters
        ----------
        entity_id: str
            ID of the entity
        entity_type: str
            Type of the entity

        Returns
        -------
        List[OverallModerationConfig]
            a list of OverallModerationConfig

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        data = unpaginate(
            cls._path,
            {"entityId": entity_id, "entityType": entity_type.value},
            cls._client,
        )
        result = next(data, None)
        if not result:
            raise errors.ClientError(status_code=404, exc_message="Not found")
        return cls.from_server_data(result)

    @classmethod
    def create(
        cls,
        timeout_sec: int,
        timeout_action: ModerationTimeoutActionType,
        entity_id: str,
        entity_type: ModerationGuardEntityType,
    ) -> OverallModerationConfig:
        """Create an OverallModerationConfig.

        .. versionadded:: v3.6

        Parameters
        ----------
        timeout_sec: int
            how long to wait for all moderation tasks in a phase to complete.
        timeout_action: ModerationTimeoutActionType
            what to do if moderation times out.
        entity_id: str
            entity, such as customModelVersion, that this configuration applies to.
        entity_type: ModerationGuardEntityType
            type of the entity defined by entity_id

        Returns
        -------
        OverallModerationConfig
            created OverallModerationConfig

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        payload = {
            "timeoutSec": timeout_sec,
            "timeoutAction": timeout_action.value,
            "entityId": entity_id,
            "entityType": entity_type.value,
        }

        response = cls._client.patch(
            cls._path,
            data=payload,
        )
        return cls.from_server_data(response.json())

    def update(
        self: OverallModerationConfig,
        timeout_sec: int,
        timeout_action: ModerationTimeoutActionType,
        entity_id: str,
        entity_type: ModerationGuardEntityType,
    ) -> None:
        """Update an OverallModerationConfig.

        .. versionadded:: v3.6

        Parameters
        ----------
        timeout_sec: int
            how long to wait for all moderation tasks in a phase to complete.
        timeout_action: ModerationTimeoutActionType
            what to do if moderation times out.
        entity_id: str
            entity, such as customModelVersion, that this configuration applies to.
        entity_type: ModerationGuardEntityType
            type of the entity defined by entity_id

        Returns
        -------
        OverallModerationConfig
            created OverallModerationConfig

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """

        payload = {
            "timeoutSec": timeout_sec,
            "timeoutAction": timeout_action.value,
            "entityId": entity_id,
            "entityType": entity_type.value,
        }

        response = self._client.patch(
            self._path,
            data=payload,
        )
        new_version = self.from_server_data(response.json())
        self._update_values(new_version)

    def refresh(self: OverallModerationConfig) -> None:
        """Update OverallModerationConfig with the latest data from server.

        .. versionadded:: v3.6

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """

        new_object = self.locate(self.entity_id, self.entity_type)
        self._update_values(new_object)



from typing import Any, Dict, List, Optional, Tuple, Union
import warnings

import trafaret as t

from datarobot._compat import Int, String

from ..utils import deprecated, deprecation_warning
from .feature_discovery import (
    DatasetDefinition,
    FeatureDiscoverySetting,
    Relationship,
    SecondaryDataset,
)
from .partitioning_methods import (
    BacktestSpecification,
    DatetimePartitioning,
    DatetimePartitioningSpecification,
    FeatureSettings,
    GroupCV,
    GroupTVH,
    Periodicity,
    RandomCV,
    RandomTVH,
    StratifiedCV,
    StratifiedTVH,
    UserCV,
    UserTVH,
)

__all__ = (
    "deprecation_warning",
    "AdvancedOptions",
    "ClassMappingAggregationSettings",
    "DatasetDefinition",
    "FeatureDiscoverySetting",
    "Relationship",
    "SecondaryDataset",
    "RandomCV",
    "StratifiedCV",
    "GroupCV",
    "UserCV",
    "RandomTVH",
    "UserTVH",
    "StratifiedTVH",
    "GroupTVH",
    "DatetimePartitioning",
    "DatetimePartitioningSpecification",
    "BacktestSpecification",
    "FeatureSettings",
    "Periodicity",
)


class AdvancedOptions(dict):  # type: ignore[type-arg]
    """
    Used when setting the target of a project to set advanced options of modeling process.

    Parameters
    ----------
    weights : Optional[str]
        The name of a column indicating the weight of each row
    response_cap : Optional[bool] or Optional[float in [0.5, 1)]
        Defaults to none here, but server defaults to False.
        If specified, it is the quantile of the response distribution to use for response capping.
    blueprint_threshold : Optional[int]
        Number of hours models are permitted to run before being excluded from later autopilot
        stages
        Minimum 1
    seed : Optional[int]
        a seed to use for randomization
    smart_downsampled : Optional[bool]
        whether to use smart downsampling to throw away excess rows of the majority class.  Only
        applicable to classification and zero-boosted regression projects.
    majority_downsampling_rate : Optional[float]
        the percentage between 0 and 100 of the majority rows that should be kept.  Specify only if
        using smart downsampling.  May not cause the majority class to become smaller than the
        minority class.
    offset : list of Optional[str]
        (New in version v2.6) the list of the names of the columns containing the offset
        of each row
    exposure : Optional[str]
        (New in version v2.6) the name of a column containing the exposure of each row
    accuracy_optimized_mb : Optional[bool]
        (New in version v2.6) Include additional, longer-running models that will be run by the
        autopilot and available to run manually.
    scaleout_modeling_mode : Optional[str]
        (Deprecated in 2.28. Will be removed in 2.30) DataRobot no longer supports scaleout models.
        Please remove any usage of this parameter as it will be removed from the API soon.
    events_count : Optional[str]
        (New in version v2.8) the name of a column specifying events count.
    monotonic_increasing_featurelist_id : Optional[str]
        (new in version 2.11) the id of the featurelist that defines the set of features
        with a monotonically increasing relationship to the target. If None,
        no such constraints are enforced. When specified, this will set a default for the project
        that can be overridden at model submission time if desired.
    monotonic_decreasing_featurelist_id : Optional[str]
        (new in version 2.11) the id of the featurelist that defines the set of features
        with a monotonically decreasing relationship to the target. If None,
        no such constraints are enforced. When specified, this will set a default for the project
        that can be overridden at model submission time if desired.
    only_include_monotonic_blueprints : Optional[bool]
        (new in version 2.11) when true, only blueprints that support enforcing
        monotonic constraints will be available in the project or selected for the autopilot.
    allowed_pairwise_interaction_groups : Optional[List[Tuple[str, ...]]]
        (New in version v2.19) For GA2M models - specify groups of columns for which pairwise
        interactions will be allowed. E.g. if set to [(A, B, C), (C, D)] then GA2M models will
        allow interactions between columns A x B, B x C, A x C, C x D. All others (A x D, B x D) will
        not be considered.
    blend_best_models: Optional[bool]
        (New in version v2.19) blend best models during Autopilot run.
    scoring_code_only: Optional[bool]
        (New in version v2.19) Keep only models that can be converted to scorable java code
        during Autopilot run
    shap_only_mode: Optional[bool]
        (New in version v2.21) Keep only models that support SHAP values during Autopilot run. Use
        SHAP-based insights wherever possible. Defaults to False.
    prepare_model_for_deployment: Optional[bool]
        (New in version v2.19) Prepare model for deployment during Autopilot run.
        The preparation includes creating reduced feature list models, retraining best model
        on higher sample size, computing insights and assigning "RECOMMENDED FOR DEPLOYMENT" label.
    consider_blenders_in_recommendation: Optional[bool]
        (New in version 2.22.0) Include blenders when selecting a model to prepare for
        deployment in an Autopilot Run. Defaults to False.
    min_secondary_validation_model_count: Optional[int]
        (New in version v2.19) Compute "All backtest" scores (datetime models) or cross validation
        scores for the specified number of the highest ranking models on the Leaderboard,
        if over the Autopilot default.
    autopilot_data_sampling_method: Optional[str]
        (New in version v2.23) one of ``datarobot.enums.DATETIME_AUTOPILOT_DATA_SAMPLING_METHOD``.
        Applicable for OTV projects only, defines if autopilot uses "random" or "latest" sampling
        when iteratively building models on various training samples. Defaults to "random" for
        duration-based projects and to "latest" for row-based projects.
    run_leakage_removed_feature_list: Optional[bool]
        (New in version v2.23) Run Autopilot on Leakage Removed feature list (if exists).
    autopilot_with_feature_discovery: Optional[bool]
        default = False
        (New in version v2.23) If true, autopilot will run on a feature list that includes features
        found via search for interactions.
    feature_discovery_supervised_feature_reduction: Optional[bool]
        (New in version v2.23) Run supervised feature reduction for feature discovery projects.
    exponentially_weighted_moving_alpha: Optional[float]
        (New in version v2.26) defaults to None, value between 0 and 1 (inclusive), indicates
        alpha parameter used in exponentially weighted moving average within feature derivation
        window.
    external_time_series_baseline_dataset_id: Optional[str]
        (New in version v2.26) If provided, will generate metrics scaled by external model
        predictions metric for time series projects. The external predictions catalog
        must be validated before autopilot starts, see
        ``Project.validate_external_time_series_baseline`` and
        :ref:`external baseline predictions documentation <external-baseline-predictions>`
        for further explanation.
    use_supervised_feature_reduction: Optional[bool]
        defaults to True,
        Time Series only. When true, during feature generation DataRobot runs a supervised
        algorithm to retain only qualifying features. Setting to false can
        severely impact autopilot duration, especially for datasets with many features.
    primary_location_column: Optional[str].
        The name of primary location column.
    protected_features: list of Optional[str].
        (New in version v2.24) A list of project features to mark as protected for
        Bias and Fairness testing calculations. Max number of protected features allowed is 10.
    preferable_target_value: Optional[str].
        (New in version v2.24) A target value that should be treated as a favorable outcome
        for the prediction. For example, if we want to check gender discrimination for
        giving a loan and our target is named ``is_bad``, then the positive outcome for
        the prediction would be ``No``, which means that the loan is good and that's
        what we treat as a favorable result for the loaner.
    fairness_metrics_set: Optional[str].
        (New in version v2.24) Metric to use for calculating fairness.
        Can be one of ``proportionalParity``, ``equalParity``, ``predictionBalance``,
        ``trueFavorableAndUnfavorableRateParity`` or
        ``favorableAndUnfavorablePredictiveValueParity``.
        Used and required only if *Bias & Fairness in AutoML* feature is enabled.
    fairness_threshold: Optional[str].
        (New in version v2.24) Threshold value for the fairness metric.
        Can be in a range of ``[0.0, 1.0]``. If the relative (i.e. normalized) fairness
        score is below the threshold, then the user will see a visual indication on the
    bias_mitigation_feature_name : Optional[str]
        The feature from protected features that will be used in a bias mitigation task to
        mitigate bias
    bias_mitigation_technique : Optional[str]
        One of datarobot.enums.BiasMitigationTechnique
        Options:
        - 'preprocessingReweighing'
        - 'postProcessingRejectionOptionBasedClassification'
        The technique by which we'll mitigate bias, which will inform which bias mitigation task
        we insert into blueprints
    include_bias_mitigation_feature_as_predictor_variable : Optional[bool]
        Whether we should also use the mitigation feature as in input to the modeler just like
        any other categorical used for training, i.e. do we want the model to "train on" this
        feature in addition to using it for bias mitigation
    default_monotonic_increasing_featurelist_id : Optional[str]
        Returned from server on Project GET request - not able to be updated by user
    default_monotonic_decreasing_featurelist_id : Optional[str]
        Returned from server on Project GET request - not able to be updated by user
    model_group_id: Optional[str] = None
        (New in version v3.3) The name of a column containing the model group id for each row.
    model_regime_id: Optional[str] = None
        (New in version v3.3) The name of a column containing the model regime id for each row.
    model_baselines: Optional[List[str]] = None
        (New in version v3.3) The list of the names of the columns containing the model baselines
    series_id: Optional[str] = None
        (New in version v3.6) The name of a column containing the series id for each row.
    forecast_distance: Optional[str] = None
        (New in version v3.6) The name of a column containing the forecast distance for each row.
    forecast_offsets: Optional[List[str]] = None
        (New in version v3.6) The list of the names of the columns containing the forecast offsets
        for each row.
    incremental_learning_only_mode: Optional[bool] = None
        (New in version v3.4) Keep only models that support incremental learning during Autopilot run.
    incremental_learning_on_best_model: Optional[bool] = None
        (New in version v3.4) Run incremental learning on the best model during Autopilot run.
    chunk_definition_id : Optional[str]
        (New in version v3.4) Unique definition for chunks needed to run automated incremental learning.
    incremental_learning_early_stopping_rounds : Optional[int] = None
        (New in version v3.4) Early stopping rounds used in the automated incremental learning service.
    number_of_incremental_learning_iterations_before_best_model_selection: Optional[int] = None
        (New in version v3.6) Number of iterations top 5 models complete prior to best model selection.
        The minimum is 1, which means no additional iterations after the first iteration (initial model) will be run.
        The maximum is 10.
    feature_engineering_prediction_point: Optional[str] = None
        (New in version v3.7) The date column to be used as the prediction point for time-based feature engineering.

    Examples
    --------
    .. code-block:: python

        import datarobot as dr
        advanced_options = dr.AdvancedOptions(
            weights='weights_column',
            offset=['offset_column'],
            exposure='exposure_column',
            response_cap=0.7,
            blueprint_threshold=2,
            smart_downsampled=True, majority_downsampling_rate=75.0)

    """

    _fields_with_defaults = {"autopilot_with_feature_discovery", "use_supervised_feature_reduction"}

    def __init__(
        self,
        weights: Optional[str] = None,
        response_cap: Optional[Union[bool, float]] = None,
        blueprint_threshold: Optional[int] = None,
        seed: Optional[int] = None,
        smart_downsampled: Optional[bool] = None,
        majority_downsampling_rate: Optional[float] = None,
        offset: Optional[List[str]] = None,
        exposure: Optional[str] = None,
        accuracy_optimized_mb: Optional[bool] = None,
        scaleout_modeling_mode: Optional[str] = None,
        events_count: Optional[str] = None,
        monotonic_increasing_featurelist_id: Optional[str] = None,
        monotonic_decreasing_featurelist_id: Optional[str] = None,
        only_include_monotonic_blueprints: Optional[bool] = None,
        allowed_pairwise_interaction_groups: Optional[List[Tuple[str, ...]]] = None,
        blend_best_models: Optional[bool] = None,
        scoring_code_only: Optional[bool] = None,
        prepare_model_for_deployment: Optional[bool] = None,
        consider_blenders_in_recommendation: Optional[bool] = None,
        min_secondary_validation_model_count: Optional[int] = None,
        shap_only_mode: Optional[bool] = None,
        autopilot_data_sampling_method: Optional[str] = None,
        run_leakage_removed_feature_list: Optional[bool] = None,
        autopilot_with_feature_discovery: Optional[bool] = False,
        feature_discovery_supervised_feature_reduction: Optional[bool] = None,
        exponentially_weighted_moving_alpha: Optional[float] = None,
        external_time_series_baseline_dataset_id: Optional[str] = None,
        use_supervised_feature_reduction: Optional[bool] = True,
        primary_location_column: Optional[str] = None,
        protected_features: Optional[List[str]] = None,
        preferable_target_value: Optional[str] = None,
        fairness_metrics_set: Optional[str] = None,
        fairness_threshold: Optional[str] = None,
        bias_mitigation_feature_name: Optional[str] = None,
        bias_mitigation_technique: Optional[str] = None,
        include_bias_mitigation_feature_as_predictor_variable: Optional[bool] = None,
        default_monotonic_increasing_featurelist_id: Optional[str] = None,
        default_monotonic_decreasing_featurelist_id: Optional[str] = None,
        model_group_id: Optional[str] = None,
        model_regime_id: Optional[str] = None,
        model_baselines: Optional[List[str]] = None,
        series_id: Optional[str] = None,
        forecast_distance: Optional[str] = None,
        forecast_offsets: Optional[List[str]] = None,
        incremental_learning_only_mode: Optional[bool] = None,
        incremental_learning_on_best_model: Optional[bool] = None,
        number_of_incremental_learning_iterations_before_best_model_selection: Optional[int] = None,
        chunk_definition_id: Optional[str] = None,
        incremental_learning_early_stopping_rounds: Optional[int] = None,
        feature_engineering_prediction_point: Optional[str] = None,
    ) -> None:
        if scaleout_modeling_mode:
            deprecation_warning(
                subject="Parameter `scaleout_modeling_mode` in Advanced Options is deprecated",
                deprecated_since_version="3.1",
                will_remove_version="3.3",
                message="DataRobot no longer supports scaleout models. "
                "Please remove this parameter.",
            )

        self.weights = weights
        self.response_cap = response_cap
        self.blueprint_threshold = blueprint_threshold
        self.seed = seed
        self.smart_downsampled = smart_downsampled
        self.majority_downsampling_rate = majority_downsampling_rate
        self.offset = offset
        self.exposure = exposure
        self.accuracy_optimized_mb = accuracy_optimized_mb
        self.events_count = events_count
        self.monotonic_increasing_featurelist_id = monotonic_increasing_featurelist_id
        self.monotonic_decreasing_featurelist_id = monotonic_decreasing_featurelist_id
        self.only_include_monotonic_blueprints = only_include_monotonic_blueprints
        self.allowed_pairwise_interaction_groups = allowed_pairwise_interaction_groups
        self.blend_best_models = blend_best_models
        self.scoring_code_only = scoring_code_only
        self.shap_only_mode = shap_only_mode
        self.prepare_model_for_deployment = prepare_model_for_deployment
        self.consider_blenders_in_recommendation = consider_blenders_in_recommendation
        self.min_secondary_validation_model_count = min_secondary_validation_model_count
        self.autopilot_data_sampling_method = autopilot_data_sampling_method
        self.run_leakage_removed_feature_list = run_leakage_removed_feature_list
        self.autopilot_with_feature_discovery = autopilot_with_feature_discovery
        self.feature_discovery_supervised_feature_reduction = (
            feature_discovery_supervised_feature_reduction
        )
        self.feature_engineering_prediction_point = feature_engineering_prediction_point
        self.exponentially_weighted_moving_alpha = exponentially_weighted_moving_alpha
        self.external_time_series_baseline_dataset_id = external_time_series_baseline_dataset_id
        self.use_supervised_feature_reduction = use_supervised_feature_reduction
        self.primary_location_column = primary_location_column
        self.protected_features = protected_features
        self.preferable_target_value = preferable_target_value
        self.fairness_metrics_set = fairness_metrics_set
        self.fairness_threshold = fairness_threshold
        self.bias_mitigation_feature_name = bias_mitigation_feature_name
        self.bias_mitigation_technique = bias_mitigation_technique
        self.include_bias_mitigation_feature_as_predictor_variable = (
            include_bias_mitigation_feature_as_predictor_variable
        )
        self.default_monotonic_increasing_featurelist_id = (
            default_monotonic_increasing_featurelist_id
        )
        self.default_monotonic_decreasing_featurelist_id = (
            default_monotonic_decreasing_featurelist_id
        )
        self.series_id = series_id
        self.forecast_distance = forecast_distance
        self.forecast_offsets = forecast_offsets

        if model_group_id:
            deprecation_warning(
                subject="Parameter `model_group_id` in Advanced Options is deprecated",
                deprecated_since_version="3.6",
                will_remove_version="3.7",
                message="Please use parameter `series_id` instead.",
            )
            if not self.series_id:
                self.series_id = model_group_id

        if model_regime_id:
            deprecation_warning(
                subject="Parameter `model_regime_id` in Advanced Options is deprecated",
                deprecated_since_version="3.6",
                will_remove_version="3.7",
                message="Please use parameter `forecast_distance` instead.",
            )
            if not self.forecast_distance:
                self.forecast_distance = model_regime_id

        if model_baselines:
            deprecation_warning(
                subject="Parameter `model_baselines` in Advanced Options is deprecated",
                deprecated_since_version="3.6",
                will_remove_version="3.7",
                message="Please us parameter `forecast_offsets` instead.",
            )
            if not self.forecast_offsets:
                self.forecast_offsets = model_baselines

        self.incremental_learning_only_mode = incremental_learning_only_mode
        self.incremental_learning_on_best_model = incremental_learning_on_best_model
        self.number_of_incremental_learning_iterations_before_best_model_selection = (
            number_of_incremental_learning_iterations_before_best_model_selection
        )
        self.chunk_definition_id = chunk_definition_id
        self.incremental_learning_early_stopping_rounds = incremental_learning_early_stopping_rounds
        attributes = {
            k: v
            for k, v in self.__dict__.items()
            if not k.startswith("__") and not callable(getattr(self, k))
        }
        super().__init__(**attributes)

    @deprecated(
        deprecated_since_version="3.3",
        will_remove_version="3.5",
        message="All advanced options returns will become AdvancedOptions objects, "
        "and the ability to use dictionary interactions will be removed.",
    )  # https://datarobot.atlassian.net/browse/DSX-3006
    def __getitem__(self, item: str) -> Any:
        if hasattr(self, item):
            return getattr(self, item)
        else:
            raise AttributeError(f"AdvancedOptions does not contain an attribute {item}.")

    @deprecated(
        deprecated_since_version="3.3",
        will_remove_version="3.5",
        message="All advanced options returns will become AdvancedOptions objects, "
        "and the ability to use dictionary interactions will be removed.",
    )  # https://datarobot.atlassian.net/browse/DSX-3006
    def __setitem__(self, key: str, value: Any) -> None:
        if hasattr(self, key):
            setattr(self, key, value)
        else:
            raise AttributeError(f"AdvancedOptions does not contain an attribute {key}.")

    @deprecated(
        deprecated_since_version="3.3",
        will_remove_version="3.5",
        message="All advanced options returns will become AdvancedOptions objects, "
        "and the ability to use dictionary interactions will be removed.",
    )  # https://datarobot.atlassian.net/browse/DSX-3006
    def get(self, __key: Any, __default: Optional[Any] = None) -> Optional[Any]:
        return super().get(__key, __default)

    @deprecated(
        deprecated_since_version="3.3",
        will_remove_version="3.5",
        message="All advanced options returns will become AdvancedOptions objects, "
        "and the ability to use dictionary interactions will be removed.",
    )  # https://datarobot.atlassian.net/browse/DSX-3006
    def pop(self, __key: Any) -> Optional[Any]:
        setattr(self, __key, None)
        return super().pop(__key)

    def update_individual_options(self, **kwargs) -> None:  # type: ignore[no-untyped-def]
        """Update individual attributes of an instance of
        :class:`AdvancedOptions <datarobot.helpers.AdvancedOptions>`.
        """
        not_updatable_attrs = set(
            [
                "default_monotonic_increasing_featurelist_id",
                "default_monotonic_decreasing_featurelist_id",
            ]
        )
        for key, val in kwargs.items():
            if hasattr(self, key) and key not in not_updatable_attrs:
                setattr(self, key, val)
            elif key in not_updatable_attrs:
                warnings.warn(
                    f"{key} is not updatable, skipping this key.",
                    stacklevel=2,
                )
            else:
                raise AttributeError(
                    f"AdvancedOptions does not contain attribute {key}. \
See `datarobot.helpers.AdvancedOptions` for all available advanced options."
                )

    def collect_payload(self) -> Dict[str, Any]:  # pylint: disable=missing-function-docstring
        payload = dict(
            weights=self.weights,
            response_cap=self.response_cap,
            blueprint_threshold=self.blueprint_threshold,
            seed=self.seed,
            majority_downsampling_rate=self.majority_downsampling_rate,
            offset=self.offset,
            exposure=self.exposure,
            accuracy_optimized_mb=self.accuracy_optimized_mb,
            events_count=self.events_count,
            monotonic_increasing_featurelist_id=self.monotonic_increasing_featurelist_id,
            monotonic_decreasing_featurelist_id=self.monotonic_decreasing_featurelist_id,
            only_include_monotonic_blueprints=self.only_include_monotonic_blueprints,
            allowed_pairwise_interaction_groups=self.allowed_pairwise_interaction_groups,
            use_supervised_feature_reduction=self.use_supervised_feature_reduction,
        )

        # Some of the optional parameters are incompatible with the others.
        # Api will return 422 if both parameters are present.
        sfd = self.feature_discovery_supervised_feature_reduction
        optional = dict(
            smart_downsampled=self.smart_downsampled,
            blend_best_models=self.blend_best_models,
            scoring_code_only=self.scoring_code_only,
            shap_only_mode=self.shap_only_mode,
            prepare_model_for_deployment=self.prepare_model_for_deployment,
            consider_blenders_in_recommendation=self.consider_blenders_in_recommendation,
            min_secondary_validation_model_count=self.min_secondary_validation_model_count,
            autopilot_data_sampling_method=self.autopilot_data_sampling_method,
            run_leakage_removed_feature_list=self.run_leakage_removed_feature_list,
            autopilot_with_feature_discovery=self.autopilot_with_feature_discovery,
            feature_discovery_supervised_feature_reduction=sfd,
            exponentially_weighted_moving_alpha=self.exponentially_weighted_moving_alpha,
            external_time_series_baseline_dataset_id=self.external_time_series_baseline_dataset_id,
            primary_location_column=self.primary_location_column,
            protected_features=self.protected_features,
            preferable_target_value=self.preferable_target_value,
            fairness_metrics_set=self.fairness_metrics_set,
            fairness_threshold=self.fairness_threshold,
            bias_mitigation_feature_name=self.bias_mitigation_feature_name,
            bias_mitigation_technique=self.bias_mitigation_technique,
            include_bias_mitigation_feature_as_predictor_variable=(
                self.include_bias_mitigation_feature_as_predictor_variable
            ),
            default_monotonic_increasing_featurelist_id=self.default_monotonic_increasing_featurelist_id,
            default_monotonic_decreasing_featurelist_id=self.default_monotonic_decreasing_featurelist_id,
            series_id=self.series_id,
            forecast_distance=self.forecast_distance,
            forecast_offsets=self.forecast_offsets,
            incremental_learning_only_mode=self.incremental_learning_only_mode,
            incremental_learning_on_best_model=self.incremental_learning_on_best_model,
            number_of_incremental_learning_iterations_before_best_model_selection=(
                self.number_of_incremental_learning_iterations_before_best_model_selection
            ),
            chunk_definition_id=self.chunk_definition_id,
            incremental_learning_early_stopping_rounds=self.incremental_learning_early_stopping_rounds,
            feature_engineering_prediction_point=self.feature_engineering_prediction_point,
        )

        payload.update({k: v for k, v in optional.items() if v is not None})

        return payload


_class_mapping_aggregation_settings_converter = t.Dict(
    {
        t.Key("max_unaggregated_class_values", optional=True): Int(),
        t.Key("min_class_support", optional=True): Int(),
        t.Key("aggregation_class_name", optional=True): String(),
        t.Key("excluded_from_aggregation", optional=True): t.List(String()),
    }
)


class ClassMappingAggregationSettings:
    """Class mapping aggregation settings.
    For multiclass projects allows fine control over which target values will be
    preserved as classes. Classes which aren't preserved will be
    - aggregated into a single "catch everything else" class in case of multiclass
    - or will be ignored in case of multilabel.
    All attributes are optional, if not specified - server side defaults will be used.

    Attributes
    ----------
    max_unaggregated_class_values : Optional[int]
        Maximum amount of unique values allowed before aggregation kicks in.
    min_class_support : Optional[int]
        Minimum number of instances necessary for each target value in the dataset.
        All values with less instances will be aggregated.
    excluded_from_aggregation : Optional[List]
        List of target values that should be guaranteed to kept as is,
        regardless of other settings.
    aggregation_class_name : Optional[str]
        If some of the values will be aggregated - this is the name of the aggregation class
        that will replace them.
    """

    def __init__(
        self,
        max_unaggregated_class_values: Optional[int] = None,
        min_class_support: Optional[int] = None,
        excluded_from_aggregation: Optional[List[str]] = None,
        aggregation_class_name: Optional[str] = None,
    ) -> None:
        self.max_unaggregated_class_values = max_unaggregated_class_values
        self.min_class_support = min_class_support
        self.excluded_from_aggregation = excluded_from_aggregation
        self.aggregation_class_name = aggregation_class_name

    def collect_payload(self) -> Dict[str, Any]:
        # pylint: disable-next=consider-using-dict-items
        return {key: self.__dict__[key] for key in self.__dict__ if self.__dict__[key] is not None}



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from enum import Enum
from typing import Any, Dict, List, Optional

from datarobot.utils import camelize


def exclude_if(
    field_name: str, field_value: Any, if_matches: Optional[List[Any]] | None = None
) -> Dict[str, Any]:
    """
    Several API fields are optional but do not accept blank or empty values.
    This function returns a dict to use with update() to construct an API payload.
    If the field value matches any value in drop_if[], it returns {}, so update() has no effect,
    and the field is not included.

    Parameters
    ----------
    field_name: name of the field. Snake case is OK; this function camelizes
    field_value: value of the field. Enums are converted to strings.
    if_matches: if value matches any in this list, exclude the field. Default is ["", {}, None].

    To intentionally include a null or blank value in the payload, modify drop_if.

    Returns
    -------
    a dict of the value (possibly empty) to update the payload with

    """
    if if_matches is None:
        if_matches = ["", {}, None]
    if field_value in if_matches:
        return {}
    else:
        string_value = field_value
        if isinstance(field_value, Enum):
            string_value = field_value.value
        return {camelize(field_name): string_value}



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from operator import itemgetter
from typing import Any, cast, Dict, Iterable, List, Optional, Union

import trafaret as t
from typing_extensions import Literal, Unpack

from datarobot._compat import Int, String
from datarobot.enums import AnomalyAssessmentStatus, DATA_SUBSET, SOURCE_TYPE
from datarobot.models.api_object import APIObject
from datarobot.models.types import (
    AnomalyAssessmentDataPoint,
    AnomalyAssessmentPreviewBin,
    AnomalyAssessmentRecordMetadata,
    RegionExplanationsData,
)
from datarobot.utils import from_api
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution

DEFAULT_BATCH_SIZE = 1000

RecordMetadataTrafaret = t.Dict(
    {
        t.Key("record_id"): String,
        t.Key("project_id"): String,
        t.Key("model_id"): String,
        t.Key("backtest"): t.Or(String(), Int),
        t.Key("source"): t.Enum(*SOURCE_TYPE.ALL),  # type: ignore[attr-defined]
        t.Key("series_id"): t.Or(String(), t.Null),
    }
)


class BaseAPIObject(APIObject):  # pylint: disable=missing-class-docstring
    def __init__(self, **record_kwargs: Unpack[AnomalyAssessmentRecordMetadata]) -> None:
        self.record_id: str = record_kwargs["record_id"]
        self.project_id: str = record_kwargs["project_id"]
        self.model_id: str = record_kwargs["model_id"]
        self.backtest: Union[str, int] = record_kwargs["backtest"]
        self.source: Union[Literal[SOURCE_TYPE.TRAINING], Literal[SOURCE_TYPE.VALIDATION]] = (
            record_kwargs["source"]
        )
        self.series_id: Optional[str] = record_kwargs["series_id"]

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(project_id={self.project_id}, model_id={self.model_id}, "
            f"series_id={self.series_id}, backtest={self.backtest}, source={self.source}, "
            f"record_id={self.record_id})"
        )

    @classmethod
    def from_server_data(
        cls,
        data: Union[Dict[str, Any], List[Dict[str, Any]]],
        keep_attrs: Optional[Iterable[str]] = None,
    ) -> "BaseAPIObject":
        """
        Instantiate an object of this class using the data directly from the server,
        meaning that the keys may have the wrong camel casing

        Parameters
        ----------
        data : dict
            The directly translated dict of JSON from the server. No casing fixes have
            taken place
        keep_attrs : list
            List of the dotted namespace notations for attributes to keep within the
            object structure even if their values are None
        """
        case_converted = from_api(data, keep_attrs=keep_attrs, keep_null_keys=True)
        return cls.from_data(case_converted)


PARAMS_TYPE = Dict[
    str,
    Union[
        int,
        str,
        Literal[DATA_SUBSET.HOLDOUT],
        Literal[SOURCE_TYPE.TRAINING],
        Literal[SOURCE_TYPE.VALIDATION],
        None,
    ],
]


class AnomalyAssessmentRecord(BaseAPIObject):
    """Object which keeps metadata about anomaly assessment insight for the particular
    subset, backtest and series and the links to proceed to get the anomaly assessment data.

    .. versionadded:: v2.25

    Attributes
    ----------
    record_id: str
        The ID of the record.
    project_id: str
        The ID of the project record belongs to.
    model_id: str
        The ID of the model record belongs to.
    backtest: int or "holdout"
        The backtest of the record.
    source: "training" or "validation"
        The source of the record
    series_id: str or None
        The series id of the record for the multiseries projects. Defined only for the multiseries
        projects.
    status: str
        The status of the insight. One of ``datarobot.enums.AnomalyAssessmentStatus``
    status_details: str
        The explanation of the status.
    start_date: str or None
        The ISO-formatted timestamp of the first prediction in the subset. Will be None if status is
        not `AnomalyAssessmentStatus.COMPLETED`.
    end_date: str or None
        The ISO-formatted timestamp of the last prediction in the subset. Will be None if status is
        not `AnomalyAssessmentStatus.COMPLETED`.
    prediction_threshold: float or None
        The threshold, all rows with anomaly scores greater or equal to it have shap explanations computed.
    preview_location: str or None
        The URL to retrieve predictions preview for the subset. Will be None if status is
        not `AnomalyAssessmentStatus.COMPLETED`.
    latest_explanations_location: str or None
        The URL to retrieve the latest predictions with the shap explanations. Will be None if status is
        not `AnomalyAssessmentStatus.COMPLETED`.
    delete_location: str
        The URL to delete anomaly assessment record and relevant insight data.
    """

    _path = "projects/{project_id}/anomalyAssessmentRecords/"
    _create_path = "projects/{project_id}/models/{model_id}/anomalyAssessmentInitialization/"

    _converter = (
        t.Dict(
            {
                t.Key("status"): t.Enum(*AnomalyAssessmentStatus.ALL),
                t.Key("status_details"): String,
                t.Key("start_date"): t.Or(String(), t.Null),
                t.Key("end_date"): t.Or(String(), t.Null),
                t.Key("prediction_threshold"): t.Or(t.Float, t.Null),
                t.Key("preview_location"): t.Or(String(), t.Null),
                t.Key("delete_location"): String(),
                t.Key("latest_explanations_location"): t.Or(String(), t.Null),
            }
        )
        .merge(RecordMetadataTrafaret)
        .ignore_extra("*")
    )

    def __init__(
        self,
        status: AnomalyAssessmentStatus,
        status_details: str,
        start_date: Optional[str],
        end_date: Optional[str],
        prediction_threshold: Optional[float],
        preview_location: Optional[str],
        delete_location: str,
        latest_explanations_location: Optional[str],
        **record_kwargs: Unpack[AnomalyAssessmentRecordMetadata],
    ) -> None:
        self.status = status
        self.status_details = status_details
        self.start_date = start_date
        self.end_date = end_date
        self.prediction_threshold = prediction_threshold
        self.preview_location = preview_location
        self.delete_location = delete_location
        self.latest_explanations_location = latest_explanations_location
        super().__init__(**record_kwargs)

    @classmethod
    def list(
        cls,
        project_id: str,
        model_id: str,
        backtest: Optional[Union[int, Literal[DATA_SUBSET.HOLDOUT]]] = None,
        source: Optional[
            Union[Literal[SOURCE_TYPE.TRAINING], Literal[SOURCE_TYPE.VALIDATION]]
        ] = None,
        series_id: Optional[str] = None,
        limit: Optional[int] = 100,
        offset: Optional[int] = 0,
        with_data_only: Optional[bool] = False,
    ) -> List["AnomalyAssessmentRecord"]:
        """Retrieve the list of the anomaly assessment records for the project and model.
        Output can be filtered and limited.

        Parameters
        ----------
        project_id: str
            The ID of the project record belongs to.
        model_id: str
            The ID of the model record belongs to.
        backtest: int or "holdout"
            The backtest to filter records by.
        source: "training" or "validation"
            The source to filter records by.
        series_id: Optional[str]
            The series id to filter records by. Can be specified for multiseries projects.
        limit: Optional[int]
            100 by default. At most this many results are returned.
        offset: Optional[int]
            This many results will be skipped.
        with_data_only: bool, False by default
            Filter by `status` == AnomalyAssessmentStatus.COMPLETED. If True, records with
            no data or not supported will be omitted.

        Returns
        -------
        AnomalyAssessmentRecord
            The anomaly assessment record.
        """
        params: PARAMS_TYPE = {"limit": limit, "offset": offset}
        if model_id:
            params["modelId"] = model_id
        if backtest:
            params["backtest"] = backtest
        if source:
            params["source"] = source
        if series_id:
            params["series_id"] = series_id
        url = cls._path.format(project_id=project_id)
        if limit == 0:  # unlimited results
            params["limit"] = DEFAULT_BATCH_SIZE
            records = cast(
                List["AnomalyAssessmentRecord"],
                [cls.from_server_data(entry) for entry in unpaginate(url, params, cls._client)],
            )
        else:
            r_data = cls._client.get(url, params=params).json()
            records = cast(
                List["AnomalyAssessmentRecord"],
                [cls.from_server_data(item) for item in r_data["data"]],
            )
        if with_data_only:
            records = [
                record for record in records if record.status == AnomalyAssessmentStatus.COMPLETED
            ]
        return records

    @classmethod
    def compute(
        cls,
        project_id: str,
        model_id: int,
        backtest: Optional[Union[int, Literal[DATA_SUBSET.HOLDOUT]]],
        source: Union[Literal[SOURCE_TYPE.TRAINING], Literal[SOURCE_TYPE.VALIDATION]],
        series_id: Optional[str] = None,
    ) -> "AnomalyAssessmentRecord":
        """Request anomaly assessment insight computation on the specified subset.

        Parameters
        ----------
        project_id: str
            The ID of the project to compute insight for.
        model_id: str
            The ID of the model to compute insight for.
        backtest: int or "holdout"
            The backtest to compute insight for.
        source: "training" or "validation"
            The source  to compute insight for.
        series_id: Optional[str]
            The series id to compute insight for. Required for multiseries projects.

        Returns
        -------
        AnomalyAssessmentRecord
            The anomaly assessment record.
        """
        payload: Dict[str, Union[int, Literal[DATA_SUBSET.HOLDOUT], str, None]] = {
            "backtest": backtest,
            "source": source,
        }
        if series_id:
            payload["series_id"] = series_id
        url = cls._create_path.format(project_id=project_id, model_id=model_id)
        response = cls._client.post(url, data=payload)
        finished_url = wait_for_async_resolution(cls._client, response.headers["Location"])
        r_data = cls._client.get(finished_url).json()
        # it ll be always one record
        return cast("AnomalyAssessmentRecord", cls.from_server_data(r_data["data"][0]))

    def delete(self) -> None:
        """Delete anomaly assessment record with preview and explanations."""
        self._client.delete(self.delete_location)

    def get_predictions_preview(self) -> AnomalyAssessmentPredictionsPreview:
        """Retrieve aggregated predictions statistics for the anomaly assessment record.

        Returns
        -------
        AnomalyAssessmentPredictionsPreview
        """
        data = self._client.get(self.preview_location).json()
        return cast(
            AnomalyAssessmentPredictionsPreview,
            AnomalyAssessmentPredictionsPreview.from_server_data(data),
        )

    def get_latest_explanations(self) -> AnomalyAssessmentExplanations:
        """Retrieve latest predictions along with shap explanations for the most anomalous records.

        Returns
        -------
        AnomalyAssessmentExplanations
        """
        data = self._client.get(self.latest_explanations_location).json()
        return cast(
            AnomalyAssessmentExplanations,
            AnomalyAssessmentExplanations.from_server_data(data),
        )

    def get_explanations(
        self,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        points_count: Optional[int] = None,
    ) -> AnomalyAssessmentExplanations:
        """Retrieve predictions along with shap explanations for the most anomalous records
        in the specified date range/for defined number of points.
        Two out of three parameters: start_date, end_date or points_count must be specified.

        Parameters
        ----------
        start_date: Optional[str]
            The start of the date range to get explanations in.
            Example: ``2020-01-01T00:00:00.000000Z``
        end_date: Optional[str]
            The end of the date range to get explanations in.
            Example: ``2020-10-01T00:00:00.000000Z``
        points_count: Optional[int]
            The number of the rows to return.

        Returns
        -------
        AnomalyAssessmentExplanations
        """
        return AnomalyAssessmentExplanations.get(
            self.project_id,
            self.record_id,
            start_date=start_date,
            end_date=end_date,
            points_count=points_count,
        )

    def get_explanations_data_in_regions(
        self, regions: List[AnomalyAssessmentPreviewBin], prediction_threshold: float = 0.0
    ) -> RegionExplanationsData:
        """Get predictions along with explanations for the specified regions, sorted by
        predictions in descending order.

        Parameters
        ----------
        regions : list of AnomalyAssessmentPreviewBin
            For each region explanations will be retrieved and merged.
        prediction_threshold : Optional[float]
            If specified, only points with score greater or equal to the threshold will be returned.

        Returns
        -------
        RegionExplanationsData
          dict in a form of {'explanations': explanations, 'shap_base_value': shap_base_value}

        """
        explanations: List[AnomalyAssessmentDataPoint] = []
        shap_base_value: Optional[float] = None
        for region in regions:
            response = self.get_explanations(
                start_date=region["start_date"], end_date=region["end_date"]
            )
            shap_base_value = response.shap_base_value
            for item in response.data:
                if item["prediction"] >= prediction_threshold:
                    explanations.append(item)
        explanations = list(sorted(explanations, key=itemgetter("prediction"), reverse=True))
        return cast(
            RegionExplanationsData,
            {"explanations": explanations, "shap_base_value": shap_base_value},
        )


class AnomalyAssessmentPredictionsPreview(BaseAPIObject):
    """Aggregated predictions over time for the corresponding anomaly assessment record.
    Intended to find the bins with highest anomaly scores.

    .. versionadded:: v2.25


    Attributes
    ----------
    record_id: str
        The ID of the record.
    project_id: str
        The ID of the project record belongs to.
    model_id: str
        The ID of the model record belongs to.
    backtest: int or "holdout"
        The backtest of the record.
    source: "training" or "validation"
        The source of the record
    series_id: str or None
        The series id of the record for the multiseries projects. Defined only for the multiseries
        projects.
    start_date: str
        the ISO-formatted timestamp of the first prediction in the subset.
    end_date: str
        the ISO-formatted timestamp of the last prediction in the subset.
    preview_bins:  list of preview_bin objects.
        The aggregated predictions for the subset.  Bins boundaries may differ from actual start/end
        dates because this is an aggregation.

    Notes
    -----

    ``PreviewBin`` contains:

    * ``start_date`` (str) : the ISO-formatted datetime of the start of the bin.
    * ``end_date`` (str) : the ISO-formatted datetime of the end of the bin.
    * ``avg_predicted`` (float or None) : the average prediction of the model in the bin. None if
      there are no entries in the bin.
    * ``max_predicted`` (float or None) : the maximum prediction of the model in the bin. None if
      there are no entries in the bin.
    * ``frequency`` (int) : the number of the rows in the bin.
    """

    _path = "projects/{project_id}/anomalyAssessmentRecords/{record_id}/predictionsPreview/"

    PreviewBinTrafaret = t.Dict(
        {
            t.Key("avg_predicted"): t.Or(t.Float(), t.Null),
            t.Key("max_predicted"): t.Or(t.Float(), t.Null),
            t.Key("start_date"): String,
            t.Key("end_date"): String,
            t.Key("frequency"): Int,
        }
    )

    _converter = (
        t.Dict(
            {
                t.Key("start_date"): String,
                t.Key("end_date"): String,
                t.Key("preview_bins"): t.List(PreviewBinTrafaret),
            }
        )
        .merge(RecordMetadataTrafaret)
        .ignore_extra("*")
    )

    def __init__(
        self,
        start_date: str,
        end_date: str,
        preview_bins: List[AnomalyAssessmentPreviewBin],
        **record_kwargs: Unpack[AnomalyAssessmentRecordMetadata],
    ) -> None:
        self.preview_bins = preview_bins
        self.start_date = start_date
        self.end_date = end_date
        super().__init__(**record_kwargs)

    @classmethod
    def get(cls, project_id: str, record_id: str) -> "AnomalyAssessmentPredictionsPreview":
        """Retrieve aggregated predictions over time.

        Parameters
        ----------
        project_id: str
            The ID of the project.
        record_id: str
            The ID of the anomaly assessment record.

        Returns
        -------
        AnomalyAssessmentPredictionsPreview

        """
        url = cls._path.format(project_id=project_id, record_id=record_id)
        r_data = cls._client.get(url).json()
        return cast("AnomalyAssessmentPredictionsPreview", cls.from_server_data(r_data))

    def find_anomalous_regions(
        self, max_prediction_threshold: float = 0.0
    ) -> List[AnomalyAssessmentPreviewBin]:
        """Sort preview bins by max_predicted value and select those with max predicted value
         greater or equal to max prediction threshold.
         Sort the result by max predicted value in descending order.

        Parameters
        ----------
        max_prediction_threshold: Optional[float]
            Return bins with maximum anomaly score greater or equal to max_prediction_threshold.

        Returns
        -------
        preview_bins: list of preview_bin
            Filtered and sorted preview bins

        """
        no_empty_bins = [bin for bin in self.preview_bins if bin["frequency"]]
        filtered_bins = [
            bin
            for bin in no_empty_bins
            if cast(float, bin["max_predicted"]) >= max_prediction_threshold
        ]
        sorted_bins = list(sorted(filtered_bins, key=itemgetter("max_predicted"), reverse=True))
        return sorted_bins


class AnomalyAssessmentExplanations(BaseAPIObject):
    """Object which keeps predictions along with shap explanations for the most anomalous records
    in the specified date range/for defined number of points.

    .. versionadded:: v2.25

    Attributes
    ----------
    record_id: str
        The ID of the record.
    project_id: str
        The ID of the project record belongs to.
    model_id: str
        The ID of the model record belongs to.
    backtest: int or "holdout"
        The backtest of the record.
    source: "training" or "validation"
        The source of the record.
    series_id: str or None
        The series id of the record for the multiseries projects. Defined only for the multiseries
        projects.
    start_date: str or None
        The ISO-formatted datetime of the first row in the ``data``. Will be None of there is no data
        in the specified range.
    end_date: str or None
        The ISO-formatted datetime of the last row in the ``data``. Will be None of there is no data
        in the specified range.
    shap_base_value: float
        Shap base value.
    count: int
        The number of points in ``data``.
    data: array of `DataPoint` objects or None
        The list of DataPoint objects in the specified date range.

    Notes
    -----

    ``DataPoint`` contains:

     * ``shap_explanation`` : None or an array of up to 10 ShapleyFeatureContribution objects.
       Only rows with the highest anomaly scores have Shapley explanations calculated.
       Value is None if prediction is lower than `prediction_threshold`.
     * ``timestamp`` (str) : ISO-formatted timestamp for the row.
     * ``prediction`` (float) : The output of the model for this row.

    ``ShapleyFeatureContribution`` contains:

     * ``feature_value`` (str) : the feature value for this row. First 50 characters are returned.
     * ``strength`` (float) : the shap value for this feature and row.
     * ``feature`` (str) : the feature name.

    """

    _path = "projects/{project_id}/anomalyAssessmentRecords/{record_id}/explanations/"

    _ShapContributionTrafaret = t.Dict(
        {t.Key("feature_value"): String, t.Key("strength"): t.Float, t.Key("feature"): String}
    )

    _RowTrafaret = t.Dict(
        {
            t.Key("shap_explanation"): t.Or(t.List(_ShapContributionTrafaret), t.Null),
            t.Key("timestamp"): String,
            t.Key("prediction"): t.Float,
        }
    )

    _converter = (
        t.Dict(
            {
                t.Key("count"): Int,
                t.Key("shap_base_value"): t.Float,
                t.Key("data"): t.List(_RowTrafaret),
                t.Key("start_date"): t.Or(String(), t.Null),
                t.Key("end_date"): t.Or(String(), t.Null),
            }
        )
        .merge(RecordMetadataTrafaret)
        .ignore_extra("*")
    )

    def __init__(
        self,
        shap_base_value: float,
        data: List[AnomalyAssessmentDataPoint],
        start_date: Optional[str],
        end_date: Optional[str],
        count: int,
        **record_kwargs: Unpack[AnomalyAssessmentRecordMetadata],
    ) -> None:
        self.shap_base_value = shap_base_value
        self.data = data
        self.count = count
        self.start_date = start_date
        self.end_date = end_date
        super().__init__(**record_kwargs)

    @classmethod
    def get(
        cls,
        project_id: str,
        record_id: str,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        points_count: Optional[int] = None,
    ) -> "AnomalyAssessmentExplanations":
        """Retrieve predictions along with shap explanations for the most anomalous records
        in the specified date range/for defined number of points.
        Two out of three parameters: start_date, end_date or points_count must be specified.

        Parameters
        ----------
        project_id: str
            The ID of the project.
        record_id: str
            The ID of the anomaly assessment record.
        start_date: Optional[str]
            The start of the date range to get explanations in.
            Example: ``2020-01-01T00:00:00.000000Z``
        end_date: Optional[str]
            The end of the date range to get explanations in.
            Example: ``2020-10-01T00:00:00.000000Z``
        points_count: Optional[int]
            The number of the rows to return.


        Returns
        -------
        AnomalyAssessmentExplanations

        """
        params: Dict[str, Union[str, int]] = {}
        if start_date:
            params["startDate"] = start_date
        if end_date:
            params["endDate"] = end_date
        if points_count:
            params["pointsCount"] = points_count

        url = cls._path.format(project_id=project_id, record_id=record_id)
        r_data = cls._client.get(url, params=params).json()
        return cast("AnomalyAssessmentExplanations", cls.from_server_data(r_data))



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from enum import Enum
from typing import List


class RemoteEventType(Enum):
    MODERATION_METRIC_CREATION_ERROR = "moderationMetricCreationError"
    MODERATION_METRIC_REPORTING_ERROR = "moderationMetricReportingError"
    MODERATION_MODEL_CONFIG_ERROR = "moderationModelConfigError"
    MODERATION_MODEL_RUNTIME_ERROR = "moderationModelRuntimeError"
    MODERATION_MODEL_SCORING_ERROR = "moderationModelScoringError"

    @classmethod
    def moderation_event_types(cls) -> List[str]:
        return [
            cls.MODERATION_METRIC_CREATION_ERROR.value,
            cls.MODERATION_METRIC_REPORTING_ERROR.value,
            cls.MODERATION_MODEL_CONFIG_ERROR.value,
            cls.MODERATION_MODEL_RUNTIME_ERROR.value,
            cls.MODERATION_MODEL_SCORING_ERROR.value,
        ]


class EventStatusType(Enum):
    """Defines the status for the MLOps event."""

    SUCCESS = "success"
    FAILURE = "failure"
    WARNING = "warning"
    INFO = "info"

    @classmethod
    def all(cls) -> List[str]:
        return [cls.SUCCESS.value, cls.FAILURE.value, cls.WARNING.value, cls.INFO.value]



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, cast, Dict, List, Optional, TYPE_CHECKING
import warnings

import trafaret as t

from datarobot._compat import String
from datarobot.utils import from_api

from ..errors import InvalidRatingTableWarning
from ..utils import get_id_from_response, recognize_sourcedata
from .api_object import APIObject

if TYPE_CHECKING:
    from datarobot.models.api_object import ServerDataDictType
    from datarobot.models.job import Job


class RatingTable(APIObject):
    """Interface to modify and download rating tables.

    Attributes
    ----------
    id : str
        The id of the rating table.
    project_id : str
        The id of the project this rating table belongs to.
    rating_table_name : str
        The name of the rating table.
    original_filename : str
        The name of the file used to create the rating table.
    parent_model_id : str
        The model id of the model the rating table was validated against.
    model_id : str
        The model id of the model that was created from the rating table.
        Can be None if a model has not been created from the rating table.
    model_job_id : str
        The id of the job to create a model from this rating table.
        Can be None if a model has not been created from the rating table.
    validation_job_id : str
        The id of the created job to validate the rating table.
        Can be None if the rating table has not been validated.
    validation_error : str
        Contains a description of any errors caused during validation.
    """

    _converter = t.Dict(
        {
            t.Key("id"): String,
            t.Key("project_id"): String,
            t.Key("rating_table_name", optional=True): String,
            t.Key("original_filename", optional=True): String,
            t.Key("parent_model_id", optional=True): String,
            t.Key("model_id", optional=True): String,
            t.Key("model_job_id", optional=True): String,
            t.Key("validation_job_id", optional=True): String,
            t.Key("validation_error", optional=True): String(allow_blank=True),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id: str,
        rating_table_name: str,
        original_filename: str,
        project_id: str,
        parent_model_id: str,
        model_id: Optional[str] = None,
        model_job_id: Optional[str] = None,
        validation_job_id: Optional[str] = None,
        validation_error: Optional[str] = None,
    ):
        self.id = id
        self.rating_table_name = rating_table_name
        self.original_filename = original_filename
        self.project_id = project_id
        self.parent_model_id = parent_model_id
        self.model_id = model_id
        self.model_job_id = model_job_id
        self.validation_job_id = validation_job_id
        self.validation_error = validation_error

    @classmethod
    def from_server_data(  # type: ignore[override] # pylint: disable=arguments-renamed
        cls,
        data: Dict[str, Any],
        should_warn: bool = True,
        keep_attrs: Optional[List[str]] = None,
    ) -> RatingTable:
        """
        Instantiate an object of this class using the data directly from the server,
        meaning that the keys may have the wrong camel casing

        Parameters
        ----------
        data : dict
            The directly translated dict of JSON from the server. No casing fixes have
            taken place
        should_warn : bool
            Whether or not to issue a warning if an invalid rating table is being retrieved.
        """
        case_converted = cast("ServerDataDictType", from_api(data, keep_attrs=keep_attrs))
        if should_warn:
            cls._warn_on_validation_error(case_converted["validation_error"])
        return cls.from_data(case_converted)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.rating_table_name})"

    @staticmethod
    def _warn_on_validation_error(validation_error: str) -> None:
        if validation_error != "":
            warnings.warn(
                "The retrieved rating table was invalid, "
                "validation error: {}".format(validation_error),
                InvalidRatingTableWarning,
                stacklevel=6,
            )

    @classmethod
    def get(cls, project_id: str, rating_table_id: str) -> RatingTable:
        """Retrieve a single rating table

        Parameters
        ----------
        project_id : str
            The ID of the project the rating table is associated with.
        rating_table_id : str
            The ID of the rating table

        Returns
        -------
        rating_table : RatingTable
            The queried instance
        """
        path = f"projects/{project_id}/ratingTables/{rating_table_id}/"
        return cls.from_location(path)

    @classmethod
    def create(
        cls,
        project_id: str,
        parent_model_id: str,
        filename: str,
        rating_table_name: str = "Uploaded Rating Table",
    ) -> Job:
        """
        Uploads and validates a new rating table CSV

        Parameters
        ----------
        project_id : str
            id of the project the rating table belongs to
        parent_model_id : str
            id of the model for which this rating table should be validated against
        filename : str
            The path of the CSV file containing the modified rating table.
        rating_table_name : Optional[str]
            A human friendly name for the new rating table. The string may be
            truncated and a suffix may be added to maintain unique names of all
            rating tables.

        Returns
        -------
        job: Job
            an instance of created async job

        Raises
        ------
        InputNotUnderstoodError
            Raised if `filename` isn't one of supported types.
        ClientError
            Raised if `parent_model_id` is invalid.
        """
        from .job import Job  # pylint: disable=import-outside-toplevel,cyclic-import

        form_data = {
            "parent_model_id": parent_model_id,
            "rating_table_name": rating_table_name,
        }
        path = f"projects/{project_id}/ratingTables/"

        default_fname = "rating_table.csv"
        filesource_kwargs = recognize_sourcedata(filename, default_fname)
        response = cls._client.build_request_with_file(
            url=path,
            form_data=form_data,
            method="post",
            file_field_name="ratingTableFile",
            **filesource_kwargs,
        )
        job_id = get_id_from_response(response)
        return Job.get(project_id, job_id)

    def download(self, filepath: str) -> None:
        """
        Download a csv file containing the contents of this rating table

        Parameters
        ----------
        filepath : str
            The path at which to save the rating table file.
        """
        path = f"projects/{self.project_id}/ratingTables/{self.id}/file"
        response = self._client.get(path)
        with open(filepath, mode="wb") as out_file:
            out_file.write(response.content)

    def rename(self, rating_table_name: str) -> None:
        """
        Renames a rating table to a different name.

        Parameters
        ----------
        rating_table_name : str
            The new name to rename the rating table to.
        """
        path = f"projects/{self.project_id}/ratingTables/{self.id}/"
        response = self._client.patch(path, data={"ratingTableName": rating_table_name})
        updated_rating_table = self.from_server_data(response.json())
        self.rating_table_name = updated_rating_table.rating_table_name

    def create_model(self) -> Job:
        """
        Creates a new model from this rating table record. This rating table
        must not already be associated with a model and must be valid.

        Returns
        -------
        job: Job
            an instance of created async job

        Raises
        ------
        ClientError
            Raised if creating model from a RatingTable that failed validation
        JobAlreadyRequested
            Raised if creating model from a RatingTable that is already
            associated with a RatingTableModel
        """
        from .model import RatingTableModel  # pylint: disable=import-outside-toplevel,cyclic-import

        return RatingTableModel.create_from_rating_table(self.project_id, self.id)



#
# Copyright 2024 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.

"""
The configuration object here applies to an individual guard.
Moderation for a model can involve multiple guards: models, NeMo guardrails, etc.

    name = dto.DtoField(t.String(max_length=MAX_GUARD_NAME_LENGTH))
    type = dto.DtoField(t.Enum(*GuardType.all()), auto_convert=True)
    description = dto.DtoField(t.String(max_length=MAX_GUARD_DESCRIPTION_LENGTH))
    created_at = dto.DtoField(RFC3339NoTZStringTrafaret(), auto_convert=True)
    creator_id = dto.DtoField(t.Or(MongoId(), t.Null()), auto_convert=True)
    # configurations are stored for a custom model (not a specific version)
    # or a playground ID
    # however, in prod, publishing the configs (to a model YAML file)
    # can be done only the latest model version (and creates a new one)
    entity_id = dto.DtoField(MongoId(), auto_convert=True)
    entity_type = dto.DtoField(t.Enum(*GuardEntityType.all()), auto_convert=True)
    deployment_id = dto.DtoFieldOptional(MongoId() | t.Null(), auto_convert=True, default=None)

"""

from __future__ import annotations

from typing import Any, Dict, List, Optional, Set

import trafaret as t
from trafaret import String

from datarobot.enums import (
    GuardType,
    ModerationGuardEntityType,
    ModerationGuardLlmType,
    ModerationGuardOotbType,
    ModerationGuardStage,
)
from datarobot.models.api_object import APIObject
from datarobot.models.moderation.intervention import GuardInterventionForConfiguration
from datarobot.models.moderation.model_info import GuardModelInfo
from datarobot.models.moderation.nemo_info import GuardNemoInfo
from datarobot.models.moderation.utils import exclude_if
from datarobot.utils import parse_time
from datarobot.utils.pagination import unpaginate


class ModerationConfiguration(APIObject):
    """Details of overall moderation configuration for a model."""

    _path = "guardConfigurations/"

    @classmethod
    def _config_path(cls, config_id: str) -> str:
        return f"{cls._path}{config_id}/"

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("name"): String(),
            t.Key("description"): String(),
            t.Key("type"): t.Enum(*[e.value for e in GuardType]),
            t.Key("entity_id"): t.String(),
            t.Key("entity_type"): t.Enum(*[e.value for e in ModerationGuardEntityType]),
            t.Key("ootb_type", optional=True): t.Enum(*[e.value for e in ModerationGuardOotbType]),
            t.Key("llm_type", optional=True): t.Enum(*[e.value for e in ModerationGuardLlmType]),
            t.Key("deployment_id", optional=True): String(allow_blank=True),
            t.Key("created_at", optional=True): t.Or(parse_time, t.Null),
            t.Key("creator_id", optional=True): String(allow_blank=True),
            t.Key("org_id", optional=True): String(allow_blank=True),
            t.Key("intervention", optional=True): GuardInterventionForConfiguration.schema,
            t.Key("model_info", optional=True): GuardModelInfo.schema,
            t.Key("nemo_info", optional=True): GuardNemoInfo.schema,
            # below are part of configuration, but not from a template
            t.Key("stages"): t.List(t.Enum(*[e.value for e in ModerationGuardStage])),
            t.Key("is_valid", optional=True): t.Bool(),
            t.Key("error_message", optional=True): String(),
            t.Key("openai_api_key", optional=True): String(),
            t.Key("openai_api_base", optional=True): String(),
            t.Key("openai_deployment_id", optional=True): String(),
            t.Key("openai_credential", optional=True): String(),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(
        self: ModerationConfiguration,
        id: str,
        name: str,
        type: GuardType,
        description: str,
        created_at: str,
        entity_id: str,
        entity_type: ModerationGuardEntityType,
        stages: list[ModerationGuardStage],
        ootb_type: Optional[str] = None,
        llm_type: Optional[str] = None,
        deployment_id: Optional[str] = None,
        creator_id: Optional[str] = None,
        org_id: Optional[str] = None,
        intervention: Optional[GuardInterventionForConfiguration | Dict[str, Any]] = None,
        model_info: Optional[GuardModelInfo | Dict[str, Any]] = None,
        nemo_info: Optional[GuardNemoInfo | Dict[str, Any]] = None,
        is_valid: Optional[bool] = True,
        error_message: Optional[str] = "",
        openai_api_key: Optional[str] = "",
        openai_api_base: Optional[str] = "",
        openai_deployment_id: Optional[str] = "",
        openai_credential: Optional[str] = "",
    ) -> None:
        self.id = id
        self.name = name
        self.type = GuardType(type)
        self.description = description
        self.created_at = created_at
        self.entity_id = entity_id
        self.entity_type = ModerationGuardEntityType(entity_type)
        self.stages = [ModerationGuardStage(x) for x in stages]
        self.creator_id = creator_id
        self.org_id = org_id
        self.nemo_info = nemo_info
        self.ootb_type = None
        if ootb_type:
            self.ootb_type = ModerationGuardOotbType(ootb_type)
        self.llm_type = None
        if llm_type:
            self.llm_type = ModerationGuardLlmType(llm_type)
        self.deployment_id = None
        if deployment_id:
            self.deployment_id = deployment_id
        # the following may arrive as objects or as dicts
        if intervention is None:
            self.intervention = None
        else:
            self.intervention = GuardInterventionForConfiguration.ensure_object(intervention)
        if model_info is None:
            self.model_info = None
        else:
            self.model_info = GuardModelInfo.ensure_object(model_info)
        self.is_valid = is_valid
        self.error_message = error_message
        self.openai_api_key = openai_api_key
        self.openai_api_base = openai_api_base
        self.openai_deployment_id = openai_deployment_id
        self.openai_credential = openai_credential

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name or self.id!r})"

    def _update_values(
        self: ModerationConfiguration, new_response: ModerationConfiguration
    ) -> None:
        # called by update() and refresh()
        fields: Set[str] = self._fields()  # type: ignore[no-untyped-call]
        for attr in fields:
            new_value = getattr(new_response, attr)
            setattr(self, attr, new_value)

    @classmethod
    def get(cls, config_id: str) -> ModerationConfiguration:
        """Get a guard configuration by ID.

        .. versionadded:: v3.6

        Parameters
        ----------
        config_id: str
            ID of the configuration

        Returns
        -------
        ModerationConfiguration
            retrieved configuration

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        path = cls._config_path(config_id)
        return cls.from_location(path)

    @classmethod
    def list(
        cls, entity_id: str, entity_type: ModerationGuardEntityType
    ) -> List[ModerationConfiguration]:
        """List Guard Configurations.

        .. versionadded:: v3.6

        Parameters
        ----------
        entity_id: str
            ID of the entity
        entity_type: ModerationGuardEntityType
            Type of the entity

        Returns
        -------
        List[ModerationConfiguration]
            a list of configurations

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        data = unpaginate(
            cls._path,
            {"entityId": entity_id, "entityType": entity_type.value},
            cls._client,
        )
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def create(
        cls,
        template_id: str,
        name: str,
        description: str,
        stages: List[ModerationGuardStage],
        entity_id: str,
        entity_type: ModerationGuardEntityType,
        intervention: Optional[GuardInterventionForConfiguration] = None,
        llm_type: Optional[ModerationGuardLlmType] = None,
        deployment_id: Optional[str] = None,
        model_info: Optional[GuardModelInfo] = None,
        nemo_info: Optional[GuardNemoInfo] = None,
        openai_api_key: Optional[str] = "",
        openai_api_base: Optional[str] = "",
        openai_deployment_id: Optional[str] = "",
        openai_credential: Optional[str] = "",
    ) -> ModerationConfiguration:
        """Create a configuration. This is not a full create from scratch; it's based on a template.

        .. versionadded:: v3.6

        Parameters
        ----------
        template_id: str
            ID of the template to base this configuration on.
        name: str
            name of the configuration.
        description: str
            description of the configuration.
        stages: List[ModerationGuardStage]
            the stages of moderation where this guard is active
        entity_id:
            ID of the custom model version or playground this configuration applies to.
        entity_type: ModerationGuardEntityType
            Type of the associated entity_id
        llm_type:
            the backing LLM this guard uses.
        nemo_info
            additional configuration for NeMo Guardrails guards.
        model_info
            additional configuration for guards using a deployed model.
        intervention
            the assessment conditions, and action the guard should take if conditions are met.
        openai_api_key: str
            Token to use for OpenAI. Deprecated; use openai_credential instead.
        openai_api_base:
            Base of the OpenAI connection
        openai_deployment_id
            ID of the OpenAI deployment
        openai_credential
            ID of the credential defined in DataRobot for OpenAI.

        Returns
        -------
        ModerationConfiguration
            created ModerationConfiguration

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        # start with required fields, then optional
        payload = {
            "templateId": template_id,
            "name": name,
            "description": description,
            "stages": stages,  # enum, but the framework converts back/forth to string
            "llmType": llm_type,
            "entityId": entity_id,
            "entityType": entity_type.value,
        }
        # omit the following fields if they're empty
        payload.update(exclude_if("deployment_id", deployment_id))
        payload.update(exclude_if("openai_api_key", openai_api_key))
        payload.update(exclude_if("openai_api_base", openai_api_base))
        payload.update(exclude_if("openai_deployment_id", openai_deployment_id))
        payload.update(exclude_if("openai_credential", openai_credential))
        payload.update(exclude_if("intervention", intervention))
        payload.update(exclude_if("model_info", model_info))
        payload.update(exclude_if("nemo_info", nemo_info))

        response = cls._client.post(
            cls._path,
            data=payload,
        )
        # todo: POST response should include entire object, but only returns ID.
        # template has the same issue.
        # todo: re-fetch until that's fixed, then use cls.from_server_data(response.json())
        return cls.get(response.json()["id"])

    def update(
        self: ModerationConfiguration,
        name: Optional[str] = None,
        description: Optional[str] = None,
        intervention: Optional[GuardInterventionForConfiguration] = None,
        llm_type: Optional[ModerationGuardLlmType] = None,
        deployment_id: Optional[str] = None,
        model_info: Optional[GuardModelInfo] = None,
        nemo_info: Optional[GuardNemoInfo] = None,
        openai_api_key: Optional[str] = "",
        openai_api_base: Optional[str] = "",
        openai_deployment_id: Optional[str] = "",
        openai_credential: Optional[str] = "",
    ) -> None:
        """Update configuration. All fields are optional, and omitted fields are left unchanged.

        entity_id, entity_type, and stages cannot be modified for a guard configuration,.

        .. versionadded:: v3.6

        Parameters
        ----------
        name: str
            name of the configuration.
        description: str
            description of the configuration.
        llm_type:
            the backing LLM this guard uses.
        nemo_info
            additional configuration for NeMo Guardrails guards.
        model_info
            additional configuration for guards using a deployed model.
        intervention
            the assessment conditions, and action the guard should take if conditions are met.
        openai_api_key: str
            Token to use for OpenAI. Deprecated; use openai_credential instead.
        openai_api_base:
            Base of the OpenAI connection
        openai_deployment_id
            ID of the OpenAI deployment
        openai_credential
            ID of the credential defined in DataRobot for OpenAI.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        payload = {}
        payload.update(exclude_if("name", name))
        payload.update(exclude_if("description", description))
        payload.update(exclude_if("llm_type", llm_type))
        payload.update(exclude_if("deployment_id", deployment_id))
        payload.update(exclude_if("openai_api_key", openai_api_key))
        payload.update(exclude_if("openai_api_base", openai_api_base))
        payload.update(exclude_if("openai_deployment_id", openai_deployment_id))
        payload.update(exclude_if("openai_credential", openai_credential))
        payload.update(exclude_if("intervention", intervention))
        payload.update(exclude_if("model_info", model_info))
        payload.update(exclude_if("nemo_info", nemo_info))

        response = self._client.patch(self._config_path(self.id), data=payload)
        data = response.json()
        new_version = self.from_server_data(data)
        self._update_values(new_version)

    def refresh(self: ModerationConfiguration) -> None:
        """Update OverallModerationConfig with the latest data from server.

        .. versionadded:: v3.6

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """

        new_object = self.get(self.id)
        self._update_values(new_object)

    def delete(self: ModerationConfiguration) -> None:
        """Delete configuration.

        .. versionadded:: v3.6

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        path = self._config_path(self.id)
        self._client.delete(path)



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

import trafaret as t
from typing_extensions import TypedDict

from datarobot.enums import (
    CategoricalStatsMethods,
    DatetimeSamplingStrategy,
    DownsamplingOperations,
    enum_to_list,
    FilterOperationFunctions,
    NumericStatsMethods,
    SamplingOperations,
    WranglingOperations,
)
from datarobot.models.api_object import APIObject


class BaseOperation(APIObject):
    """Single base transformation unit in Data Wrangler recipe."""

    def __init__(self, directive: str, arguments: Any):
        self.directive = directive
        self.arguments = arguments


class WranglingOperation(BaseOperation):
    _converter = t.Dict(
        {
            t.Key("directive"): t.Enum(*enum_to_list(WranglingOperations)),
            t.Key("arguments"): t.Mapping(t.String(), t.Any()),
        }
    ).allow_extra("*")


class DownsamplingOperation(BaseOperation):
    _converter = t.Dict(
        {
            t.Key("directive"): t.Enum(*enum_to_list(DownsamplingOperations)),
            t.Key("arguments"): t.Mapping(t.String(), t.Any()),
        }
    ).allow_extra("*")


class SamplingOperation(BaseOperation):
    _converter = t.Dict(
        {
            t.Key("directive"): t.Enum(*enum_to_list(SamplingOperations)),
            t.Key("arguments"): t.Mapping(t.String(), t.Any()),
        }
    ).allow_extra("*")


class BaseTimeAwareTask(APIObject):
    def __init__(self, name: str, arguments: Dict[str, Any]):
        self.name = name
        self.arguments = arguments


class TaskPlanElement(APIObject):
    def __init__(self, column: str, task_list: List[BaseTimeAwareTask]):
        self.column = column
        self.task_list = task_list


class CategoricalStats(BaseTimeAwareTask):
    def __init__(self, methods: List[CategoricalStatsMethods], window_size: int):
        super().__init__("categorical-stats", {"window_size": window_size, "methods": methods})


class NumericStats(BaseTimeAwareTask):
    def __init__(self, methods: List[NumericStatsMethods], window_size: int):
        super().__init__("numeric-stats", {"window_size": window_size, "methods": methods})


class Lags(BaseTimeAwareTask):
    def __init__(self, orders: List[int]):
        super().__init__("lags", {"orders": orders})


class LagsOperation(WranglingOperation):
    """
    Generate lags in a window.
    """

    def __init__(
        self,
        column: str,
        orders: List[int],
        datetime_partition_column: str,
        multiseries_id_column: Optional[str] = None,
    ):
        super().__init__(
            directive=WranglingOperations.LAGS,
            arguments={
                "column": column,
                "orders": orders,
                "datetime_partition_column": datetime_partition_column,
                "multiseries_id_column": multiseries_id_column,
            },
        )


class WindowCategoricalStatsOperation(WranglingOperation):
    """
    Generate rolling statistics in a window for categorical features.
    """

    def __init__(
        self,
        column: str,
        window_size: int,
        methods: List[CategoricalStatsMethods],
        datetime_partition_column: str,
        multiseries_id_column: Optional[str] = None,
        rolling_most_frequent_udf: Optional[str] = None,
    ):
        super().__init__(
            directive=WranglingOperations.WINDOW_CATEGORICAL_STATS,
            arguments={
                "column": column,
                "window_size": window_size,
                "methods": methods,
                "datetime_partition_column": datetime_partition_column,
                "multiseries_id_column": multiseries_id_column,
                "rolling_most_frequent_user_defined_function": rolling_most_frequent_udf,
            },
        )


class WindowNumericStatsOperation(WranglingOperation):
    """Generate various rolling numeric statistics in a window. Output could be a several columns."""

    def __init__(
        self,
        column: str,
        window_size: int,
        methods: List[NumericStatsMethods],
        datetime_partition_column: str,
        multiseries_id_column: Optional[str] = None,
        rolling_median_udf: Optional[str] = None,
    ):
        super().__init__(
            directive=WranglingOperations.WINDOW_NUMERIC_STATS,
            arguments={
                "column": column,
                "window_size": window_size,
                "methods": methods,
                "datetime_partition_column": datetime_partition_column,
                "multiseries_id_column": multiseries_id_column,
                "rolling_median_user_defined_function": rolling_median_udf,
            },
        )


class TimeSeriesOperation(WranglingOperation):
    """Operation to generate a dataset ready for time series modeling: with forecast point, forecast distances,
    known in advance columns, etc.
    """

    def __init__(
        self,
        target_column: str,
        datetime_partition_column: str,
        forecast_distances: List[int],
        task_plan: List[TaskPlanElement],
        baseline_periods: Optional[List[int]] = None,
        known_in_advance_columns: Optional[List[str]] = None,
        multiseries_id_column: Optional[str] = None,
        rolling_median_udf: Optional[str] = None,
        rolling_most_frequent_udf: Optional[str] = None,
        forecast_point: Optional[datetime] = None,
    ):
        """

        Parameters
        ----------
        target_column
        datetime_partition_column
        forecast_distances
        task_plan:
            contains a task list for each column
        baseline_periods:
            generates naive features from the target. For example: baseline period = 1 corresponds to the naive
            latest baseline.
        known_in_advance_columns
        multiseries_id_column
        rolling_median_udf:
            Fully qualified path to rolling median user defined function. Used to optimize sql execution with snowflake.
        rolling_most_frequent_udf:
            Fully qualified path to rolling most frequent user defined function.
        forecast_point:
             To use at prediction time.
        """
        arguments = {
            "target_column": target_column,
            "datetime_partition_column": datetime_partition_column,
            "forecast_distances": forecast_distances,
            "task_plan": task_plan,
            "multiseries_id_column": multiseries_id_column,
            "known_in_advance_columns": known_in_advance_columns,
            "baseline_periods": baseline_periods,
            "rolling_median_user_defined_function": rolling_median_udf,
            "rolling_most_frequent_user_defined_function": rolling_most_frequent_udf,
            "forecast_point": forecast_point,
        }
        super().__init__(directive=WranglingOperations.TIME_SERIES, arguments=arguments)


class ComputeNewOperation(WranglingOperation):
    def __init__(self, expression: str, new_feature_name: str):
        super().__init__(
            directive=WranglingOperations.COMPUTE_NEW,
            arguments={"expression": expression, "new_feature_name": new_feature_name},
        )


class RenameColumnsOperation(WranglingOperation):
    def __init__(self, column_mappings: Dict[str, str]):
        """
        column_mapping: dict, where
            key:  str
                Original name
            value: str
                New name
        """
        super().__init__(
            directive=WranglingOperations.RENAME_COLUMNS,
            arguments={
                "column_mappings": [
                    {"original_name": k, "new_name": v} for k, v in column_mappings.items()
                ]
            },
        )


class FilterCondition(TypedDict):
    column: str
    function: FilterOperationFunctions
    function_arguments: List[Union[str, int, float]]


class FilterOperation(WranglingOperation):
    """Filter rows."""

    def __init__(
        self,
        conditions: List[FilterCondition],
        keep_rows: Optional[bool] = True,
        operator: Optional[str] = "and",
    ):
        """
        keep_rows: bool
            If matching rows should be kept or dropped
        operator: str
            "and" or "or"
        conditions: list of FilterCondition

        """
        super().__init__(
            directive=WranglingOperations.FILTER,
            arguments={"keep_rows": keep_rows, "operator": operator, "conditions": conditions},
        )


class DropColumnsOperation(WranglingOperation):
    def __init__(self, columns: List[str]):
        """
        columns:
            Columns to delete
        """
        super().__init__(
            directive=WranglingOperations.DROP_COLUMNS,
            arguments={"columns": columns},
        )


class RandomSamplingOperation(SamplingOperation):
    def __init__(self, rows: int, seed: Optional[int] = None):
        super().__init__(
            directive=SamplingOperations.RANDOM_SAMPLE,
            arguments={"rows": rows, "seed": seed},
        )


class DatetimeSamplingOperation(SamplingOperation):
    def __init__(
        self,
        datetime_partition_column: str,
        rows: int,
        strategy: Optional[Union[str, DatetimeSamplingStrategy]] = None,
        multiseries_id_column: Optional[str] = None,
        selected_series: Optional[List[str]] = None,
    ):
        super().__init__(
            directive=SamplingOperations.DATETIME_SAMPLE,
            arguments={
                "rows": rows,
                "strategy": strategy,
                "datetime_partition_column": datetime_partition_column,
                "multiseries_id_column": multiseries_id_column,
                "selected_series": selected_series,
            },
        )



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datetime import datetime
from typing import List, Optional, TYPE_CHECKING

import trafaret as t

from datarobot._compat import Int, String
from datarobot.models.api_object import APIObject
from datarobot.utils import parse_time

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    class DataQualityWarning(TypedDict):
        has_kia_missing_values_in_forecast_window: bool
        insufficient_rows_for_evaluating_models: bool
        single_class_actual_value_column: bool

    class DetectedActualValueColumn(TypedDict):
        missing_count: int
        name: str


class PredictionDataset(APIObject):
    """A dataset uploaded to make predictions

    Typically created via `project.upload_dataset`

    Attributes
    ----------
    id : str
        the id of the dataset
    project_id : str
        the id of the project the dataset belongs to
    created : str
        the time the dataset was created
    name : str
        the name of the dataset
    num_rows : int
        the number of rows in the dataset
    num_columns : int
        the number of columns in the dataset
    forecast_point : datetime.datetime or None
        For time series projects only. This is the default point relative to which predictions will
        be generated, based on the forecast window of the project.  See the time series
        :ref:`predictions documentation <time-series-predict>` for more information.
    predictions_start_date : datetime.datetime or None, optional
        For time series projects only. The start date for bulk predictions. Note that this
        parameter is for generating historical predictions using the training data. This parameter
        should be provided in conjunction with ``predictions_end_date``. Can't be provided with the
        ``forecast_point`` parameter.
    predictions_end_date : datetime.datetime or None, optional
        For time series projects only. The end date for bulk predictions, exclusive. Note that this
        parameter is for generating historical predictions using the training data. This parameter
        should be provided in conjunction with ``predictions_start_date``. Can't be provided with
        the ``forecast_point`` parameter.
    relax_known_in_advance_features_check : Optional[bool]
        (New in version v2.15) For time series projects only. If True, missing values in the
        known in advance features are allowed in the forecast window at the prediction time.
        If omitted or False, missing values are not allowed.
    data_quality_warnings : dict, optional
        (New in version v2.15) A dictionary that contains available warnings about potential
        problems in this prediction dataset. Available warnings include:

         - has_kia_missing_values_in_forecast_window (`bool`)
            Applicable for time series projects. If True, known in advance features
            have missing values in forecast window which may decrease prediction accuracy.
         - insufficient_rows_for_evaluating_models  (`bool`)
            Applicable for datasets which are used as external test sets. If True, there is not
            enough rows in dataset to calculate insights.
         - single_class_actual_value_column (`bool`)
            Applicable for datasets which are used as external test sets. If True, actual value
            column has only one class and such insights as ROC curve can not be calculated.
            Only applies for binary classification projects or unsupervised projects.

    forecast_point_range : list[datetime.datetime] or None, optional
        (New in version v2.20) For time series projects only. Specifies the range of dates available
        for use as a forecast point.
    data_start_date : datetime.datetime or None, optional
        (New in version v2.20) For time series projects only. The minimum primary date of this
        prediction dataset.
    data_end_date : datetime.datetime or None, optional
        (New in version v2.20) For time series projects only. The maximum primary date of this
        prediction dataset.
    max_forecast_date : datetime.datetime or None, optional
        (New in version v2.20) For time series projects only. The maximum forecast date of this
        prediction dataset.
    actual_value_column : string, optional
        (New in version v2.21) Optional, only available for unsupervised projects,
        in case dataset was uploaded with actual value column specified. Name of the
        column which will be used to calculate the classification metrics and insights.
    detected_actual_value_columns : list of dict, optional
        (New in version v2.21) For unsupervised projects only, list of detected actual value
        columns information containing missing count and name for each column.
    contains_target_values : Optional[bool]
        (New in version v2.21)  Only for supervised projects. If True, dataset contains target
        values and can be used to calculate the classification metrics and insights.
    secondary_datasets_config_id: string or None, optional
        (New in version v2.23) The Id of the alternative secondary dataset config
        to use during prediction for Feature discovery project.
    """

    _path_template = "projects/{}/predictionDatasets/{}/"

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("project_id"): String(),
            t.Key("created"): String(),
            t.Key("name"): String(),
            t.Key("num_rows"): Int(),
            t.Key("num_columns"): Int(),
            t.Key("forecast_point", optional=True): parse_time,
            t.Key("predictions_start_date", optional=True): parse_time,
            t.Key("predictions_end_date", optional=True): parse_time,
            t.Key("relax_known_in_advance_features_check", optional=True): t.Bool(),
            # do not forget to update `test_data_quality_warnings`
            # in datarobot-python-api-tests repo
            t.Key("data_quality_warnings", optional=True): t.Dict(
                {
                    t.Key("has_kia_missing_values_in_forecast_window"): t.Bool(),
                    t.Key("insufficient_rows_for_evaluating_models"): t.Bool(),
                    t.Key("single_class_actual_value_column"): t.Bool(),
                }
            ).allow_extra("*"),
            t.Key("forecast_point_range", optional=True): t.List(parse_time),
            t.Key("data_start_date", optional=True): parse_time,
            t.Key("data_end_date", optional=True): parse_time,
            t.Key("max_forecast_date", optional=True): parse_time,
            t.Key("actual_value_column", optional=True): String(),
            t.Key("detected_actual_value_columns", optional=True): t.List(
                t.Dict({t.Key("missing_count"): Int(), t.Key("name"): String()}).ignore_extra("*")
            ),
            t.Key("contains_target_values", optional=True): t.Bool(),
            t.Key("secondary_datasets_config_id", optional=True): String(),
        }
    ).allow_extra("*")

    def __init__(
        self,
        project_id: str,
        id: str,
        name: str,
        created: str,
        num_rows: int,
        num_columns: int,
        forecast_point: Optional[datetime] = None,
        predictions_start_date: Optional[datetime] = None,
        predictions_end_date: Optional[datetime] = None,
        relax_known_in_advance_features_check: Optional[bool] = None,
        data_quality_warnings: Optional[DataQualityWarning] = None,
        forecast_point_range: Optional[List[datetime]] = None,
        data_start_date: Optional[datetime] = None,
        data_end_date: Optional[datetime] = None,
        max_forecast_date: Optional[datetime] = None,
        actual_value_column: Optional[str] = None,
        detected_actual_value_columns: Optional[List[DetectedActualValueColumn]] = None,
        contains_target_values: Optional[bool] = None,
        secondary_datasets_config_id: Optional[str] = None,
    ) -> None:
        self.project_id = project_id
        self.id = id
        self.name = name
        self.created = created
        self.num_rows = num_rows
        self.num_columns = num_columns
        self.forecast_point = forecast_point
        self.predictions_start_date = predictions_start_date
        self.predictions_end_date = predictions_end_date
        self.relax_known_in_advance_features_check = relax_known_in_advance_features_check
        self.data_quality_warnings = data_quality_warnings
        self.forecast_point_range = forecast_point_range
        self.data_start_date = data_start_date
        self.data_end_date = data_end_date
        self.max_forecast_date = max_forecast_date
        self.detected_actual_value_columns = detected_actual_value_columns
        self.actual_value_column = actual_value_column
        self.contains_target_values = contains_target_values
        self._path = self._path_template.format(project_id, id)
        self.secondary_datasets_config_id = secondary_datasets_config_id

    def __repr__(self) -> str:
        return f"PredictionDataset({self.name!r})"

    @classmethod
    def get(cls, project_id: str, dataset_id: str) -> PredictionDataset:
        """
        Retrieve information about a dataset uploaded for predictions

        Parameters
        ----------
        project_id:
            the id of the project to query
        dataset_id:
            the id of the dataset to retrieve

        Returns
        -------
        dataset: PredictionDataset
            A dataset uploaded to make predictions
        """
        path = cls._path_template.format(project_id, dataset_id)
        return cls.from_location(path)

    def delete(self) -> None:
        """Delete a dataset uploaded for predictions

        Will also delete predictions made using this dataset and cancel any predict jobs using
        this dataset.
        """
        self._client.delete(self._path)



# flake8: noqa
from .chat import Chat
from .chat_prompt import ChatPrompt
from .comparison_chat import ComparisonChat
from .comparison_prompt import ComparisonPrompt
from .cost_metric_configurations import CostMetricConfiguration
from .custom_model_embedding_validation import CustomModelEmbeddingValidation
from .custom_model_llm_validation import CustomModelLLMValidation
from .evaluation_dataset_configuration import EvaluationDatasetConfiguration
from .evaluation_dataset_metric_aggregation import EvaluationDatasetMetricAggregation
from .insights_configuration import InsightsConfiguration
from .llm import LLMDefinition
from .llm_blueprint import LLMBlueprint
from .llm_test_configuration import LLMTestConfiguration
from .llm_test_result import LLMTestResult
from .metric_insights import MetricInsights
from .nemo_configuration import NemoConfiguration
from .ootb_metric_configuration import PlaygroundOOTBMetricConfiguration
from .playground import Playground
from .playground_moderation_configuration import Intervention, ModerationConfigurationWithoutId
from .prompt_trace import PromptTrace
from .sidecar_model_metric import SidecarModelMetricValidation
from .synthetic_evaluation_dataset_generation import SyntheticEvaluationDataset
from .user_limits import UserLimits
from .vector_database import CustomModelVectorDatabaseValidation, VectorDatabase
